from __future__ import absolute_import

from django.test.client import RequestFactory
from django.utils import timezone
from rest_framework import serializers
from rest_framework.fields import SkipField
from rest_framework.reverse import reverse

from rest_framework_json_api.exceptions import Conflict
from rest_framework_json_api.relations import (
    HyperlinkedRelatedField,
    ResourceRelatedField,
    SerializerMethodHyperlinkedRelatedField
)
from rest_framework_json_api.utils import format_resource_type

from . import TestBase
from example.models import Author, Blog, Comment, Entry
from example.serializers import CommentSerializer
from example.views import EntryViewSet


class TestResourceRelatedField(TestBase):

    def setUp(self):
        super(TestResourceRelatedField, self).setUp()
        self.blog = Blog.objects.create(name='Some Blog', tagline="It's a blog")
        self.entry = Entry.objects.create(
            blog=self.blog,
            headline='headline',
            body_text='body_text',
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=3
        )
        for i in range(1, 6):
            name = 'some_author{}'.format(i)
            self.entry.authors.add(
                Author.objects.create(name=name, email='{}@example.org'.format(name))
            )

        self.comment = Comment.objects.create(
            entry=self.entry,
            body='testing one two three',
            author=Author.objects.first()
        )

    def test_data_in_correct_format_when_instantiated_with_blog_object(self):
        serializer = BlogFKSerializer(instance={'blog': self.blog})

        expected_data = {
            'type': format_resource_type('Blog'),
            'id': str(self.blog.id)
        }

        actual_data = serializer.data['blog']

        self.assertEqual(actual_data, expected_data)

    def test_data_in_correct_format_when_instantiated_with_entry_object(self):
        serializer = EntryFKSerializer(instance={'entry': self.entry})

        expected_data = {
            'type': format_resource_type('Entry'),
            'id': str(self.entry.id)
        }

        actual_data = serializer.data['entry']

        self.assertEqual(actual_data, expected_data)

    def test_deserialize_primitive_data_blog(self):
        serializer = BlogFKSerializer(data={
            'blog': {
                'type': format_resource_type('Blog'),
                'id': str(self.blog.id)
            }
        }
        )

        self.assertTrue(serializer.is_valid())
        self.assertEqual(serializer.validated_data['blog'], self.blog)

    def test_validation_fails_for_wrong_type(self):
        with self.assertRaises(Conflict) as cm:
            serializer = BlogFKSerializer(data={
                'blog': {
                    'type': 'Entries',
                    'id': str(self.blog.id)
                }
            }
            )
            serializer.is_valid()
        the_exception = cm.exception
        self.assertEqual(the_exception.status_code, 409)

    def test_serialize_many_to_many_relation(self):
        serializer = EntryModelSerializer(instance=self.entry)

        type_string = format_resource_type('Author')
        author_pks = Author.objects.values_list('pk', flat=True)
        expected_data = [{'type': type_string, 'id': str(pk)} for pk in author_pks]

        self.assertEqual(
            serializer.data['authors'],
            expected_data
        )

    def test_deserialize_many_to_many_relation(self):
        type_string = format_resource_type('Author')
        author_pks = Author.objects.values_list('pk', flat=True)
        authors = [{'type': type_string, 'id': pk} for pk in author_pks]

        serializer = EntryModelSerializer(data={'authors': authors, 'comments': []})

        self.assertTrue(serializer.is_valid())
        self.assertEqual(len(serializer.validated_data['authors']), Author.objects.count())
        for author in serializer.validated_data['authors']:
            self.assertIsInstance(author, Author)

    def test_read_only(self):
        serializer = EntryModelSerializer(
            data={'authors': [], 'comments': [{'type': 'Comments', 'id': 2}]}
        )
        serializer.is_valid(raise_exception=True)
        self.assertNotIn('comments', serializer.validated_data)

    def test_invalid_resource_id_object(self):
        comment = {'body': 'testing 123', 'entry': {'type': 'entry'}, 'author': {'id': '5'}}
        serializer = CommentSerializer(data=comment)
        assert not serializer.is_valid()
        assert serializer.errors == {
            'author': ["Invalid resource identifier object: missing 'type' attribute"],
            'entry': ["Invalid resource identifier object: missing 'id' attribute"]
        }


class TestHyperlinkedFieldBase(TestBase):

    def setUp(self):
        super(TestHyperlinkedFieldBase, self).setUp()
        self.blog = Blog.objects.create(name='Some Blog', tagline="It's a blog")
        self.entry = Entry.objects.create(
            blog=self.blog,
            headline='headline',
            body_text='body_text',
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=3
        )
        self.comment = Comment.objects.create(
            entry=self.entry,
            body='testing one two three',
        )

        self.request = RequestFactory().get(reverse('entry-detail', kwargs={'pk': self.entry.pk}))
        self.view = EntryViewSet(request=self.request, kwargs={'entry_pk': self.entry.id})


class TestHyperlinkedRelatedField(TestHyperlinkedFieldBase):

    def test_single_hyperlinked_related_field(self):
        field = HyperlinkedRelatedField(
            related_link_view_name='entry-blog',
            related_link_url_kwarg='entry_pk',
            self_link_view_name='entry-relationships',
            read_only=True,
        )
        field._context = {'request': self.request, 'view': self.view}
        field.field_name = 'blog'

        self.assertRaises(NotImplementedError, field.to_representation, self.entry)
        self.assertRaises(SkipField, field.get_attribute, self.entry)

        links_expected = {
            'self': 'http://testserver/entries/{}/relationships/blog'.format(self.entry.pk),
            'related': 'http://testserver/entries/{}/blog'.format(self.entry.pk)
        }
        got = field.get_links(self.entry)
        self.assertEqual(got, links_expected)

    def test_many_hyperlinked_related_field(self):
        field = HyperlinkedRelatedField(
            related_link_view_name='entry-comments',
            related_link_url_kwarg='entry_pk',
            self_link_view_name='entry-relationships',
            read_only=True,
            many=True
        )
        field._context = {'request': self.request, 'view': self.view}
        field.field_name = 'comments'

        self.assertRaises(NotImplementedError, field.to_representation, self.entry.comments.all())
        self.assertRaises(SkipField, field.get_attribute, self.entry)

        links_expected = {
            'self': 'http://testserver/entries/{}/relationships/comments'.format(self.entry.pk),
            'related': 'http://testserver/entries/{}/comments'.format(self.entry.pk)
        }
        got = field.child_relation.get_links(self.entry)
        self.assertEqual(got, links_expected)


class TestSerializerMethodHyperlinkedRelatedField(TestHyperlinkedFieldBase):

    def test_single_serializer_method_hyperlinked_related_field(self):
        serializer = EntryModelSerializerWithHyperLinks(
            instance=self.entry,
            context={
                'request': self.request,
                'view': self.view
            }
        )
        field = serializer.fields['blog']

        self.assertRaises(NotImplementedError, field.to_representation, self.entry)
        self.assertRaises(SkipField, field.get_attribute, self.entry)

        expected = {
            'self': 'http://testserver/entries/{}/relationships/blog'.format(self.entry.pk),
            'related': 'http://testserver/entries/{}/blog'.format(self.entry.pk)
        }
        got = field.get_links(self.entry)
        self.assertEqual(got, expected)

    def test_many_serializer_method_hyperlinked_related_field(self):
        serializer = EntryModelSerializerWithHyperLinks(
            instance=self.entry,
            context={
                'request': self.request,
                'view': self.view
            }
        )
        field = serializer.fields['comments']

        self.assertRaises(NotImplementedError, field.to_representation, self.entry)
        self.assertRaises(SkipField, field.get_attribute, self.entry)

        expected = {
            'self': 'http://testserver/entries/{}/relationships/comments'.format(self.entry.pk),
            'related': 'http://testserver/entries/{}/comments'.format(self.entry.pk)
        }
        got = field.get_links(self.entry)
        self.assertEqual(got, expected)

    def test_get_blog(self):
        serializer = EntryModelSerializerWithHyperLinks(instance=self.entry)
        got = serializer.get_blog(self.entry)
        expected = self.entry.blog

        self.assertEqual(got, expected)

    def test_get_comments(self):
        serializer = EntryModelSerializerWithHyperLinks(instance=self.entry)
        got = serializer.get_comments(self.entry)
        expected = self.entry.comments.all()

        self.assertListEqual(list(got), list(expected))


class BlogResourceRelatedField(ResourceRelatedField):
    def get_queryset(self):
        return Blog.objects


class BlogFKSerializer(serializers.Serializer):
    blog = BlogResourceRelatedField()


class EntryFKSerializer(serializers.Serializer):
    entry = ResourceRelatedField(queryset=Entry.objects)


class EntryModelSerializer(serializers.ModelSerializer):
    authors = ResourceRelatedField(many=True, queryset=Author.objects)
    comments = ResourceRelatedField(many=True, read_only=True)

    class Meta:
        model = Entry
        fields = ('authors', 'comments')


class EntryModelSerializerWithHyperLinks(serializers.ModelSerializer):
    blog = SerializerMethodHyperlinkedRelatedField(
        related_link_view_name='entry-blog',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        many=True,
    )
    comments = SerializerMethodHyperlinkedRelatedField(
        related_link_view_name='entry-comments',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        many=True,
    )

    class Meta:
        model = Entry
        fields = ('blog', 'comments',)

    def get_blog(self, obj):
        return obj.blog

    def get_comments(self, obj):
        return obj.comments.all()



from django.utils import timezone
from rest_framework.test import APITestCase

from example.factories import CommentFactory
from example.models import Author, Blog, Comment, Entry


class PerformanceTestCase(APITestCase):
    def setUp(self):
        self.author = Author.objects.create(name='Super powerful superhero', email='i.am@lost.com')
        self.blog = Blog.objects.create(name='Some Blog', tagline="It's a blog")
        self.other_blog = Blog.objects.create(name='Other blog', tagline="It's another blog")
        self.first_entry = Entry.objects.create(
            blog=self.blog,
            headline='headline one',
            body_text='body_text two',
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=3
        )
        self.second_entry = Entry.objects.create(
            blog=self.blog,
            headline='headline two',
            body_text='body_text one',
            pub_date=timezone.now(),
            mod_date=timezone.now(),
            n_comments=0,
            n_pingbacks=0,
            rating=1
        )
        self.comment = Comment.objects.create(entry=self.first_entry)
        CommentFactory.create_batch(50)

    def test_query_count_no_includes(self):
        """ We expect a simple list view to issue only two queries.

        1. The number of results in the set (e.g. a COUNT query),
           only necessary because we're using PageNumberPagination
        2. The SELECT query for the set
        """
        with self.assertNumQueries(2):
            response = self.client.get('/comments?page[size]=25')
            self.assertEqual(len(response.data['results']), 25)

    def test_query_count_include_author(self):
        """ We expect a list view with an include have three queries:

        1. Primary resource COUNT query
        2. Primary resource SELECT
        3. Authors prefetched
        4. Author types prefetched
        5. Entries prefetched
        """
        with self.assertNumQueries(5):
            response = self.client.get('/comments?include=author&page[size]=25')
            self.assertEqual(len(response.data['results']), 25)

    def test_query_select_related_entry(self):
        """ We expect a list view with an include have two queries:

        1. Primary resource COUNT query
        2. Primary resource SELECT + SELECT RELATED writer(author) and bio
        """
        with self.assertNumQueries(2):
            response = self.client.get('/comments?include=writer&page[size]=25')
            self.assertEqual(len(response.data['results']), 25)



from rest_framework.reverse import reverse
from rest_framework.test import APITestCase

from example.models import Blog, Entry


class DJATestFilters(APITestCase):
    """
    tests of JSON:API filter backends
    """
    fixtures = ('blogentry',)

    def setUp(self):
        self.entries = Entry.objects.all()
        self.blogs = Blog.objects.all()
        self.url = reverse('nopage-entry-list')
        self.fs_url = reverse('filterset-entry-list')
        self.no_fs_url = reverse('nofilterset-entry-list')

    def test_sort(self):
        """
        test sort
        """
        response = self.client.get(self.url, data={'sort': 'headline'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        headlines = [c['attributes']['headline'] for c in dja_response['data']]
        sorted_headlines = sorted(headlines)
        self.assertEqual(headlines, sorted_headlines)

    def test_sort_reverse(self):
        """
        confirm switching the sort order actually works
        """
        response = self.client.get(self.url, data={'sort': '-headline'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        headlines = [c['attributes']['headline'] for c in dja_response['data']]
        sorted_headlines = sorted(headlines)
        self.assertNotEqual(headlines, sorted_headlines)

    def test_sort_double_negative(self):
        """
        what if they provide multiple `-`'s? It's OK.
        """
        response = self.client.get(self.url, data={'sort': '--headline'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        headlines = [c['attributes']['headline'] for c in dja_response['data']]
        sorted_headlines = sorted(headlines)
        self.assertNotEqual(headlines, sorted_headlines)

    def test_sort_invalid(self):
        """
        test sort of invalid field
        """
        response = self.client.get(self.url,
                                   data={'sort': 'nonesuch,headline,-not_a_field'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid sort parameters: nonesuch,-not_a_field")

    def test_sort_camelcase(self):
        """
        test sort of camelcase field name
        """
        response = self.client.get(self.url, data={'sort': 'bodyText'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        blog_ids = [(c['attributes']['bodyText'] or '') for c in dja_response['data']]
        sorted_blog_ids = sorted(blog_ids)
        self.assertEqual(blog_ids, sorted_blog_ids)

    def test_sort_underscore(self):
        """
        test sort of underscore field name
        Do we allow this notation in a search even if camelcase is in effect?
        "Be conservative in what you send, be liberal in what you accept"
                   --  https://en.wikipedia.org/wiki/Robustness_principle
        """
        response = self.client.get(self.url, data={'sort': 'body_text'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        blog_ids = [(c['attributes']['bodyText'] or '') for c in dja_response['data']]
        sorted_blog_ids = sorted(blog_ids)
        self.assertEqual(blog_ids, sorted_blog_ids)

    def test_sort_related(self):
        """
        test sort via related field using jsonapi path `.` and django orm `__` notation.
        ORM relations must be predefined in the View's .ordering_fields attr
        """
        for datum in ('blog__id', 'blog.id'):
            response = self.client.get(self.url, data={'sort': datum})
            self.assertEqual(response.status_code, 200,
                             msg=response.content.decode("utf-8"))
            dja_response = response.json()
            blog_ids = [c['relationships']['blog']['data']['id'] for c in dja_response['data']]
            sorted_blog_ids = sorted(blog_ids)
            self.assertEqual(blog_ids, sorted_blog_ids)

    def test_filter_exact(self):
        """
        filter for an exact match
        """
        response = self.client.get(self.url, data={'filter[headline]': 'CHEM3271X'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(len(dja_response['data']), 1)

    def test_filter_exact_fail(self):
        """
        failed search for an exact match
        """
        response = self.client.get(self.url, data={'filter[headline]': 'XXXXX'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(len(dja_response['data']), 0)

    def test_filter_isnull(self):
        """
        search for null value
        """
        response = self.client.get(self.url, data={'filter[bodyText.isnull]': 'true'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(
            len(dja_response['data']),
            len([k for k in self.entries if k.body_text is None])
        )

    def test_filter_not_null(self):
        """
        search for not null
        """
        response = self.client.get(self.url, data={'filter[bodyText.isnull]': 'false'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(
            len(dja_response['data']),
            len([k for k in self.entries if k.body_text is not None])
        )

    def test_filter_isempty(self):
        """
        search for an empty value (different from null!)
        the easiest way to do this is search for r'^$'
        """
        response = self.client.get(self.url, data={'filter[bodyText.regex]': '^$'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(len(dja_response['data']),
                         len([k for k in self.entries
                              if k.body_text is not None and
                              len(k.body_text) == 0]))

    def test_filter_related(self):
        """
        filter via a relationship chain
        """
        response = self.client.get(self.url, data={'filter[blog.name]': 'ANTB'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(len(dja_response['data']),
                         len([k for k in self.entries
                              if k.blog.name == 'ANTB']))

    def test_filter_related_fieldset_class(self):
        """
        filter via a FilterSet class instead of filterset_fields shortcut
        This tests a shortcut for a longer ORM path: `bname` is a shortcut
        name for `blog.name`.
        """
        response = self.client.get(self.fs_url, data={'filter[bname]': 'ANTB'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(len(dja_response['data']),
                         len([k for k in self.entries
                              if k.blog.name == 'ANTB']))

    def test_filter_related_missing_fieldset_class(self):
        """
        filter via with neither filterset_fields nor filterset_class
        This should return an error for any filter[]
        """
        response = self.client.get(self.no_fs_url, data={'filter[bname]': 'ANTB'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid filter[bname]")

    def test_filter_fields_union_list(self):
        """
        test field for a list of values(ORed): ?filter[field.in]': 'val1,val2,val3
        """
        response = self.client.get(self.url,
                                   data={'filter[headline.in]': 'CLCV2442V,XXX,BIOL3594X'})
        dja_response = response.json()
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        self.assertEqual(
            len(dja_response['data']),
            len([k for k in self.entries if k.headline == 'CLCV2442V']) +
            len([k for k in self.entries if k.headline == 'XXX']) +
            len([k for k in self.entries if k.headline == 'BIOL3594X']),
            msg="filter field list (union)")

    def test_filter_fields_intersection(self):
        """
        test fields (ANDed): ?filter[field1]': 'val1&filter[field2]'='val2
        """
        #
        response = self.client.get(self.url,
                                   data={'filter[headline.regex]': '^A',
                                         'filter[body_text.icontains]': 'in'})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertGreater(len(dja_response['data']), 1)
        self.assertEqual(
            len(dja_response['data']),
            len([k for k in self.entries if k.headline.startswith('A') and
                 'in' in k.body_text.lower()]))

    def test_filter_invalid_association_name(self):
        """
        test for filter with invalid filter association name
        """
        response = self.client.get(self.url, data={'filter[nonesuch]': 'CHEM3271X'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid filter[nonesuch]")

    def test_filter_empty_association_name(self):
        """
        test for filter with missing association name
        error texts are different depending on whether QueryParameterValidationFilter is in use.
        """
        response = self.client.get(self.url, data={'filter[]': 'foobar'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'], "invalid query parameter: filter[]")

    def test_filter_no_brackets(self):
        """
        test for `filter=foobar` with missing filter[association] name
        """
        response = self.client.get(self.url, data={'filter': 'foobar'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid query parameter: filter")

    def test_filter_missing_right_bracket(self):
        """
        test for filter missing right bracket
        """
        response = self.client.get(self.url, data={'filter[headline': 'foobar'})
        self.assertEqual(response.status_code, 400, msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid query parameter: filter[headline")

    def test_filter_no_brackets_rvalue(self):
        """
        test for `filter=` with missing filter[association] and value
        """
        response = self.client.get(self.url + '?filter=')
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid query parameter: filter")

    def test_filter_no_brackets_equal(self):
        """
        test for `filter` with missing filter[association] name and =value
        """
        response = self.client.get(self.url + '?filter')
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid query parameter: filter")

    def test_filter_malformed_left_bracket(self):
        """
        test for filter with invalid filter syntax
        """
        response = self.client.get(self.url, data={'filter[': 'foobar'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'], "invalid query parameter: filter[")

    def test_filter_missing_rvalue(self):
        """
        test for filter with missing value to test against
        this should probably be an error rather than ignoring the filter:
        https://django-filter.readthedocs.io/en/latest/guide/tips.html#filtering-by-an-empty-string
        """
        response = self.client.get(self.url, data={'filter[headline]': ''})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "missing value for query parameter filter[headline]")

    def test_filter_missing_rvalue_equal(self):
        """
        test for filter with missing value to test against
        this should probably be an error rather than ignoring the filter:
        """
        response = self.client.get(self.url + '?filter[headline]')
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "missing value for query parameter filter[headline]")

    def test_filter_single_relation(self):
        """
        test for filter with a single relation
        e.g. filterset-entries?filter[authors.id]=1
        looks for entries written by (at least) author.id=1
        """
        response = self.client.get(self.fs_url, data={'filter[authors.id]': 1})

        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()

        ids = [k['id'] for k in dja_response['data']]

        expected_ids = [str(k.id) for k in self.entries.filter(authors__id=1)]

        self.assertEqual(set(ids), set(expected_ids))

    def test_filter_repeated_relations(self):
        """
        test for filters with repeated relations
        e.g. filterset-entries?filter[authors.id]=1&filter[authors.id]=2
        looks for entries written by (at least) author.id=1 AND author.id=2
        """
        response = self.client.get(self.fs_url, data={'filter[authors.id]': [1, 2]})

        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()

        ids = [k['id'] for k in dja_response['data']]

        expected_ids = [str(k.id) for k in self.entries.filter(authors__id=1).filter(authors__id=2)]

        self.assertEqual(set(ids), set(expected_ids))

    def test_filter_in(self):
        """
        test for the in filter
        e.g. filterset-entries?filter[authors.id.in]=1,2
        looks for entries written by (at least) author.id=1 OR author.id=2
        """
        response = self.client.get(self.fs_url, data={'filter[authors.id.in]': '1,2'})

        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()

        ids = [k['id'] for k in dja_response['data']]

        expected_ids = [str(k.id) for k in self.entries.filter(authors__id__in=[1, 2])]

        self.assertEqual(set(ids), set(expected_ids))

    def test_search_keywords(self):
        """
        test for `filter[search]="keywords"` where some of the keywords are in the entry and
        others are in the related blog.
        """
        response = self.client.get(self.url, data={'filter[search]': 'barnard field research'})
        expected_result = {
            'data': [
                {
                    'type': 'posts',
                    'id': '7',
                    'attributes': {
                        'headline': 'ANTH3868X',
                        'bodyText': 'ETHNOGRAPHIC FIELD RESEARCH IN NYC',
                        'pubDate': None,
                        'modDate': None},
                    'relationships': {
                        'blog': {
                            'data': {
                                'type': 'blogs',
                                'id': '1'
                            }
                        },
                        'blogHyperlinked': {
                            'links': {
                                'self': 'http://testserver/entries/7/relationships/blog_hyperlinked',  # noqa: E501
                                'related': 'http://testserver/entries/7/blog'}
                        },
                        'authors': {
                            'meta': {
                                'count': 0
                            },
                            'data': []
                        },
                        'comments': {
                            'meta': {
                                'count': 0
                            },
                            'data': []
                        },
                        'commentsHyperlinked': {
                            'links': {
                                'self': 'http://testserver/entries/7/relationships/comments_hyperlinked',  # noqa: E501
                                'related': 'http://testserver/entries/7/comments'
                            }
                        },
                        'suggested': {
                            'links': {
                                'self': 'http://testserver/entries/7/relationships/suggested',
                                'related': 'http://testserver/entries/7/suggested/'
                            },
                            'data': [
                                {'type': 'entries', 'id': '1'},
                                {'type': 'entries', 'id': '2'},
                                {'type': 'entries', 'id': '3'},
                                {'type': 'entries', 'id': '4'},
                                {'type': 'entries', 'id': '5'},
                                {'type': 'entries', 'id': '6'},
                                {'type': 'entries', 'id': '8'},
                                {'type': 'entries', 'id': '9'},
                                {'type': 'entries', 'id': '10'},
                                {'type': 'entries', 'id': '11'},
                                {'type': 'entries', 'id': '12'}
                            ]
                        },
                        'suggestedHyperlinked': {
                            'links': {
                                'self': 'http://testserver/entries/7/relationships/suggested_hyperlinked',  # noqa: E501
                                'related': 'http://testserver/entries/7/suggested/'}
                        },
                        'tags': {'data': [], 'meta': {'count': 0}},
                        'featuredHyperlinked': {
                            'links': {
                                'self': 'http://testserver/entries/7/relationships/featured_hyperlinked',  # noqa: E501
                                'related': 'http://testserver/entries/7/featured'
                            }
                        }
                    },
                    'meta': {
                        'bodyFormat': 'text'
                    }
                }
            ]
        }
        assert response.json() == expected_result

    def test_search_multiple_keywords(self):
        """
        test for `filter[search]=keyword1...` (keyword1 [AND keyword2...])

        See the four search_fields defined in views.py which demonstrate both searching
        direct fields (entry) and following ORM links to related fields (blog):
            `search_fields = ('headline', 'body_text', 'blog__name', 'blog__tagline')`

        SearchFilter searches for items that match all whitespace separated keywords across
        the many fields.

        This code tests that functionality by comparing the result of the GET request
        with the equivalent results used by filtering the test data via the model manager.
        To do so, iterate over the list of given searches:
        1. For each keyword, search the 4 search_fields for a match and then get the result
           set which is the union of all results for the given keyword.
        2. Intersect those results sets such that *all* keywords are represented.
        See `example/fixtures/blogentry.json` for the test content that the searches are based on.
        The searches test for both direct entries and related blogs across multiple fields.
        """
        for searches in ("research", "chemistry", "nonesuch",
                         "research seminar", "research nonesuch",
                         "barnard classic", "barnard ethnographic field research"):
            response = self.client.get(self.url, data={'filter[search]': searches})
            self.assertEqual(response.status_code, 200, msg=response.content.decode("utf-8"))
            dja_response = response.json()
            keys = searches.split()
            # dicts keyed by the search keys for the 4 search_fields:
            headline = {}      # list of entry ids where key is in entry__headline
            body_text = {}     # list of entry ids where key is in entry__body_text
            blog_name = {}     # list of entry ids where key is in entry__blog__name
            blog_tagline = {}  # list of entry ids where key is in entry__blog__tagline
            for key in keys:
                headline[key] = [str(k.id) for k in
                                 self.entries.filter(headline__icontains=key)]
                body_text[key] = [str(k.id) for k in
                                  self.entries.filter(body_text__icontains=key)]
                blog_name[key] = [str(k.id) for k in
                                  self.entries.filter(blog__name__icontains=key)]
                blog_tagline[key] = [str(k.id) for k in
                                     self.entries.filter(blog__tagline__icontains=key)]
            union = []  # each list item is a set of entry ids matching the given key
            for key in keys:
                union.append(set(headline[key] + body_text[key] +
                                 blog_name[key] + blog_tagline[key]))
            # all keywords must be present: intersect the keyword sets
            expected_ids = set.intersection(*union)
            expected_len = len(expected_ids)
            self.assertEqual(len(dja_response['data']), expected_len)
            returned_ids = set([k['id'] for k in dja_response['data']])
            self.assertEqual(returned_ids, expected_ids)

    def test_param_invalid(self):
        """
        Test a "wrong" query parameter
        """
        response = self.client.get(self.url, data={'garbage': 'foo'})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "invalid query parameter: garbage")

    def test_param_duplicate_sort(self):
        """
        Test a duplicated query parameter:
        `?sort=headline&page[size]=3&sort=bodyText` is not allowed.
        This is not so obvious when using a data dict....
        """
        response = self.client.get(self.url,
                                   data={'sort': ['headline', 'bodyText'],
                                         'page[size]': 3}
                                   )
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "repeated query parameter not allowed: sort")

    def test_param_duplicate_page(self):
        """
        test a duplicated page[size] query parameter
        """
        response = self.client.get(self.fs_url, data={'page[size]': [1, 2]})
        self.assertEqual(response.status_code, 400,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(dja_response['errors'][0]['detail'],
                         "repeated query parameter not allowed: page[size]")

    def test_many_params(self):
        """
        Test that filter params aren't ignored when many params are present
        """
        response = self.client.get(self.url,
                                   data={'filter[headline.regex]': '^A',
                                         'filter[body_text.regex]': '^IN',
                                         'filter[blog.name]': 'ANTB',
                                         'page[size]': 3})
        self.assertEqual(response.status_code, 200,
                         msg=response.content.decode("utf-8"))
        dja_response = response.json()
        self.assertEqual(len(dja_response['data']), 1)
        self.assertEqual(dja_response['data'][0]['id'], '1')



import pytest
from django.contrib.auth import get_user_model
from django.test import override_settings
from rest_framework import serializers
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.views import APIView

from rest_framework_json_api import utils
from rest_framework_json_api.utils import get_included_serializers

from example.serializers import AuthorSerializer, BlogSerializer, CommentSerializer, EntrySerializer

pytestmark = pytest.mark.django_db


class NonModelResourceSerializer(serializers.Serializer):
    class Meta:
        resource_name = 'users'


class ResourceSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ('username',)
        model = get_user_model()


def test_get_resource_name():
    view = APIView()
    context = {'view': view}
    with override_settings(JSON_API_FORMAT_TYPES=None):
        assert 'APIViews' == utils.get_resource_name(context), 'not formatted'

    context = {'view': view}
    with override_settings(JSON_API_FORMAT_TYPES='dasherize'):
        assert 'api-views' == utils.get_resource_name(context), 'derived from view'

    view.model = get_user_model()
    assert 'users' == utils.get_resource_name(context), 'derived from view model'

    view.resource_name = 'custom'
    assert 'custom' == utils.get_resource_name(context), 'manually set on view'

    view.response = Response(status=403)
    assert 'errors' == utils.get_resource_name(context), 'handles 4xx error'

    view.response = Response(status=500)
    assert 'errors' == utils.get_resource_name(context), 'handles 500 error'

    view = GenericAPIView()
    view.serializer_class = ResourceSerializer
    context = {'view': view}
    assert 'users' == utils.get_resource_name(context), 'derived from serializer'

    view.serializer_class.Meta.resource_name = 'rcustom'
    assert 'rcustom' == utils.get_resource_name(context), 'set on serializer'

    view = GenericAPIView()
    view.serializer_class = NonModelResourceSerializer
    context = {'view': view}
    assert 'users' == utils.get_resource_name(context), 'derived from non-model serializer'


@pytest.mark.parametrize("format_type,output", [
    ('camelize', {'fullName': {'last-name': 'a', 'first-name': 'b'}}),
    ('capitalize', {'FullName': {'last-name': 'a', 'first-name': 'b'}}),
    ('dasherize', {'full-name': {'last-name': 'a', 'first-name': 'b'}}),
    ('underscore', {'full_name': {'last-name': 'a', 'first-name': 'b'}}),
])
def test_format_field_names(settings, format_type, output):
    settings.JSON_API_FORMAT_FIELD_NAMES = format_type

    value = {'full_name': {'last-name': 'a', 'first-name': 'b'}}
    assert utils.format_field_names(value, format_type) == output


def test_format_value():
    assert utils.format_value('first_name', 'camelize') == 'firstName'
    assert utils.format_value('first_name', 'capitalize') == 'FirstName'
    assert utils.format_value('first_name', 'dasherize') == 'first-name'
    assert utils.format_value('first-name', 'underscore') == 'first_name'


def test_format_resource_type():
    assert utils.format_resource_type('first_name', 'capitalize') == 'FirstNames'
    assert utils.format_resource_type('first_name', 'camelize') == 'firstNames'


class SerializerWithIncludedSerializers(EntrySerializer):
    included_serializers = {
        'blog': BlogSerializer,
        'authors': 'example.serializers.AuthorSerializer',
        'comments': 'example.serializers.CommentSerializer',
        'self': 'self'  # this wouldn't make sense in practice (and would be prohibited by
        # IncludedResourcesValidationMixin) but it's useful for the test
    }


def test_get_included_serializers_against_class():
    klass = SerializerWithIncludedSerializers
    included_serializers = get_included_serializers(klass)
    expected_included_serializers = {
        'blog': BlogSerializer,
        'authors': AuthorSerializer,
        'comments': CommentSerializer,
        'self': klass
    }
    assert included_serializers.keys() == klass.included_serializers.keys(), (
        'the keys must be preserved'
    )

    assert included_serializers == expected_included_serializers


def test_get_included_serializers_against_instance():
    klass = SerializerWithIncludedSerializers
    instance = klass()
    included_serializers = get_included_serializers(instance)
    expected_included_serializers = {
        'blog': BlogSerializer,
        'authors': AuthorSerializer,
        'comments': CommentSerializer,
        'self': klass
    }
    assert included_serializers.keys() == klass.included_serializers.keys(), (
        'the keys must be preserved'
    )

    assert included_serializers == expected_included_serializers



"""
Renderers
"""
import copy
from collections import OrderedDict, defaultdict
from collections.abc import Iterable

import inflection
from django.db.models import Manager
from django.utils import encoding
from rest_framework import relations, renderers
from rest_framework.fields import SkipField, get_attribute
from rest_framework.relations import PKOnlyObject
from rest_framework.serializers import BaseSerializer, ListSerializer, Serializer
from rest_framework.settings import api_settings

import rest_framework_json_api
from rest_framework_json_api import utils
from rest_framework_json_api.relations import HyperlinkedMixin, ResourceRelatedField, SkipDataMixin
from rest_framework_json_api.settings import json_api_settings


class JSONRenderer(renderers.JSONRenderer):
    """
    The `JSONRenderer` exposes a number of methods that you may override if you need highly
    custom rendering control.

    Render a JSON response per the JSON API spec:

    .. code-block:: json

        {
          "data": [
            {
              "type": "companies",
              "id": 1,
              "attributes": {
                "name": "Mozilla",
                "slug": "mozilla",
                "date-created": "2014-03-13 16:33:37"
              }
            }
          ]
        }
    """

    media_type = 'application/vnd.api+json'
    format = 'vnd.api+json'

    @classmethod
    def extract_attributes(cls, fields, resource):
        """
        Builds the `attributes` object of the JSON API resource object.
        """
        data = OrderedDict()
        render_nested_as_attribute = json_api_settings.SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE
        for field_name, field in iter(fields.items()):
            # ID is always provided in the root of JSON API so remove it from attributes
            if field_name == 'id':
                continue
            # don't output a key for write only fields
            if fields[field_name].write_only:
                continue
            # Skip fields with relations
            if isinstance(
                    field, (relations.RelatedField, relations.ManyRelatedField)
            ):
                continue

            if isinstance(field, BaseSerializer) and not render_nested_as_attribute:
                continue

            # Skip read_only attribute fields when `resource` is an empty
            # serializer. Prevents the "Raw Data" form of the browsable API
            # from rendering `"foo": null` for read only fields
            try:
                resource[field_name]
            except KeyError:
                if fields[field_name].read_only:
                    continue

            data.update({
                field_name: resource.get(field_name)
            })

        return utils.format_field_names(data)

    @classmethod
    def extract_relationships(cls, fields, resource, resource_instance):
        """
        Builds the relationships top level object based on related serializers.
        """
        # Avoid circular deps
        from rest_framework_json_api.relations import ResourceRelatedField

        data = OrderedDict()
        render_nested_as_attribute = json_api_settings.SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE

        # Don't try to extract relationships from a non-existent resource
        if resource_instance is None:
            return

        for field_name, field in iter(fields.items()):
            # Skip URL field
            if field_name == api_settings.URL_FIELD_NAME:
                continue

            # don't output a key for write only fields
            if fields[field_name].write_only:
                continue

            # Skip fields without relations
            if not isinstance(
                field, (relations.RelatedField, relations.ManyRelatedField, BaseSerializer)
            ):
                continue

            if isinstance(field, BaseSerializer) and render_nested_as_attribute:
                continue

            source = field.source
            relation_type = utils.get_related_resource_type(field)

            if isinstance(field, relations.HyperlinkedIdentityField):
                resolved, relation_instance = utils.get_relation_instance(
                    resource_instance, source, field.parent
                )
                if not resolved:
                    continue
                # special case for HyperlinkedIdentityField
                relation_data = list()

                # Don't try to query an empty relation
                relation_queryset = relation_instance \
                    if relation_instance is not None else list()

                for related_object in relation_queryset:
                    relation_data.append(
                        OrderedDict([
                            ('type', relation_type),
                            ('id', encoding.force_str(related_object.pk))
                        ])
                    )

                data.update({field_name: {
                    'links': {
                        "related": resource.get(field_name)},
                    'data': relation_data,
                    'meta': {
                        'count': len(relation_data)
                    }
                }})
                continue

            relation_data = {}
            if isinstance(field, HyperlinkedMixin):
                field_links = field.get_links(resource_instance, field.related_link_lookup_field)
                relation_data.update({'links': field_links} if field_links else dict())
                data.update({field_name: relation_data})

            if isinstance(field, (ResourceRelatedField, )):
                if not isinstance(field, SkipDataMixin):
                    relation_data.update({'data': resource.get(field_name)})

                data.update({field_name: relation_data})
                continue

            if isinstance(
                    field, (relations.PrimaryKeyRelatedField, relations.HyperlinkedRelatedField)
            ):
                resolved, relation = utils.get_relation_instance(
                    resource_instance, '%s_id' % source, field.parent
                )
                if not resolved:
                    continue
                relation_id = relation if resource.get(field_name) else None
                relation_data = {
                    'data': (
                        OrderedDict([
                            ('type', relation_type), ('id', encoding.force_str(relation_id))
                        ])
                        if relation_id is not None else None)
                }

                if (
                    isinstance(field, relations.HyperlinkedRelatedField) and
                    resource.get(field_name)
                ):
                    relation_data.update(
                        {
                            'links': {
                                'related': resource.get(field_name)
                            }
                        }
                    )
                data.update({field_name: relation_data})
                continue

            if isinstance(field, relations.ManyRelatedField):
                resolved, relation_instance = utils.get_relation_instance(
                    resource_instance, source, field.parent
                )
                if not resolved:
                    continue

                relation_data = {}

                if isinstance(resource.get(field_name), Iterable):
                    relation_data.update(
                        {
                            'meta': {'count': len(resource.get(field_name))}
                        }
                    )

                if isinstance(field.child_relation, ResourceRelatedField):
                    # special case for ResourceRelatedField
                    relation_data.update(
                        {'data': resource.get(field_name)}
                    )

                if isinstance(field.child_relation, HyperlinkedMixin):
                    field_links = field.child_relation.get_links(
                        resource_instance,
                        field.child_relation.related_link_lookup_field
                    )
                    relation_data.update(
                        {'links': field_links}
                        if field_links else dict()
                    )

                    data.update({field_name: relation_data})
                    continue

                relation_data = list()
                for nested_resource_instance in relation_instance:
                    nested_resource_instance_type = (
                        relation_type or
                        utils.get_resource_type_from_instance(nested_resource_instance)
                    )

                    relation_data.append(OrderedDict([
                        ('type', nested_resource_instance_type),
                        ('id', encoding.force_str(nested_resource_instance.pk))
                    ]))
                data.update({
                    field_name: {
                        'data': relation_data,
                        'meta': {
                            'count': len(relation_data)
                        }
                    }
                })
                continue

            if isinstance(field, ListSerializer):
                resolved, relation_instance = utils.get_relation_instance(
                    resource_instance, source, field.parent
                )
                if not resolved:
                    continue

                relation_data = list()

                serializer_data = resource.get(field_name)
                resource_instance_queryset = list(relation_instance)
                if isinstance(serializer_data, list):
                    for position in range(len(serializer_data)):
                        nested_resource_instance = resource_instance_queryset[position]
                        nested_resource_instance_type = (
                            relation_type or
                            utils.get_resource_type_from_instance(nested_resource_instance)
                        )

                        relation_data.append(OrderedDict([
                            ('type', nested_resource_instance_type),
                            ('id', encoding.force_str(nested_resource_instance.pk))
                        ]))

                    data.update({field_name: {'data': relation_data}})
                    continue

            if isinstance(field, Serializer):
                relation_instance_id = getattr(resource_instance, source + "_id", None)
                if not relation_instance_id:
                    resolved, relation_instance = utils.get_relation_instance(
                        resource_instance, source, field.parent
                    )
                    if not resolved:
                        continue

                    if relation_instance is not None:
                        relation_instance_id = relation_instance.pk

                data.update({
                    field_name: {
                        'data': (
                            OrderedDict([
                                ('type', relation_type),
                                ('id', encoding.force_str(relation_instance_id))
                            ]) if resource.get(field_name) else None)
                    }
                })
                continue

        return utils.format_field_names(data)

    @classmethod
    def extract_relation_instance(cls, field, resource_instance):
        """
        Determines what instance represents given relation and extracts it.

        Relation instance is determined exactly same way as it determined
        in parent serializer
        """
        try:
            res = field.get_attribute(resource_instance)
            if isinstance(res, PKOnlyObject):
                return get_attribute(resource_instance, field.source_attrs)
            return res
        except SkipField:
            return None

    @classmethod
    def extract_included(cls, fields, resource, resource_instance, included_resources,
                         included_cache):
        """
        Adds related data to the top level included key when the request includes
        ?include=example,example_field2
        """
        # this function may be called with an empty record (example: Browsable Interface)
        if not resource_instance:
            return

        current_serializer = fields.serializer
        context = current_serializer.context
        included_serializers = utils.get_included_serializers(current_serializer)
        included_resources = copy.copy(included_resources)
        included_resources = [inflection.underscore(value) for value in included_resources]
        render_nested_as_attribute = json_api_settings.SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE

        for field_name, field in iter(fields.items()):
            # Skip URL field
            if field_name == api_settings.URL_FIELD_NAME:
                continue

            # Skip fields without relations
            if not isinstance(
                field, (relations.RelatedField, relations.ManyRelatedField, BaseSerializer)
            ):
                continue

            if isinstance(field, BaseSerializer) and render_nested_as_attribute:
                continue

            try:
                included_resources.remove(field_name)
            except ValueError:
                # Skip fields not in requested included resources
                # If no child field, directly continue with the next field
                if field_name not in [node.split('.')[0] for node in included_resources]:
                    continue

            relation_instance = cls.extract_relation_instance(
                field, resource_instance
            )
            if isinstance(relation_instance, Manager):
                relation_instance = relation_instance.all()

            serializer_data = resource.get(field_name)

            if isinstance(field, relations.ManyRelatedField):
                serializer_class = included_serializers[field_name]
                field = serializer_class(relation_instance, many=True, context=context)
                serializer_data = field.data

            if isinstance(field, relations.RelatedField):
                if relation_instance is None or not serializer_data:
                    continue

                many = field._kwargs.get('child_relation', None) is not None

                if isinstance(field, ResourceRelatedField) and not many:
                    already_included = serializer_data['type'] in included_cache and \
                        serializer_data['id'] in included_cache[serializer_data['type']]

                    if already_included:
                        continue

                serializer_class = included_serializers[field_name]
                field = serializer_class(relation_instance, many=many, context=context)
                serializer_data = field.data

            new_included_resources = [key.replace('%s.' % field_name, '', 1)
                                      for key in included_resources
                                      if field_name == key.split('.')[0]]

            if isinstance(field, ListSerializer):
                serializer = field.child
                relation_type = utils.get_resource_type_from_serializer(serializer)
                relation_queryset = list(relation_instance)

                if serializer_data:
                    for position in range(len(serializer_data)):
                        serializer_resource = serializer_data[position]
                        nested_resource_instance = relation_queryset[position]
                        resource_type = (
                            relation_type or
                            utils.get_resource_type_from_instance(nested_resource_instance)
                        )
                        serializer_fields = utils.get_serializer_fields(
                            serializer.__class__(
                                nested_resource_instance, context=serializer.context
                            )
                        )
                        new_item = cls.build_json_resource_obj(
                            serializer_fields,
                            serializer_resource,
                            nested_resource_instance,
                            resource_type,
                            getattr(serializer, '_poly_force_type_resolution', False)
                        )
                        included_cache[new_item['type']][new_item['id']] = \
                            utils.format_field_names(new_item)
                        cls.extract_included(
                            serializer_fields,
                            serializer_resource,
                            nested_resource_instance,
                            new_included_resources,
                            included_cache,
                        )

            if isinstance(field, Serializer):
                relation_type = utils.get_resource_type_from_serializer(field)

                # Get the serializer fields
                serializer_fields = utils.get_serializer_fields(field)
                if serializer_data:
                    new_item = cls.build_json_resource_obj(
                        serializer_fields,
                        serializer_data,
                        relation_instance,
                        relation_type,
                        getattr(field, '_poly_force_type_resolution', False)
                    )
                    included_cache[new_item['type']][new_item['id']] = utils.format_field_names(
                        new_item
                    )
                    cls.extract_included(
                        serializer_fields,
                        serializer_data,
                        relation_instance,
                        new_included_resources,
                        included_cache,
                    )

    @classmethod
    def extract_meta(cls, serializer, resource):
        """
        Gathers the data from serializer fields specified in meta_fields and adds it to
        the meta object.
        """
        if hasattr(serializer, 'child'):
            meta = getattr(serializer.child, 'Meta', None)
        else:
            meta = getattr(serializer, 'Meta', None)
        meta_fields = getattr(meta, 'meta_fields', [])
        data = OrderedDict()
        for field_name in meta_fields:
            data.update({
                field_name: resource.get(field_name)
            })
        return data

    @classmethod
    def extract_root_meta(cls, serializer, resource):
        """
        Calls a `get_root_meta` function on a serializer, if it exists.
        """
        many = False
        if hasattr(serializer, 'child'):
            many = True
            serializer = serializer.child

        data = {}
        if getattr(serializer, 'get_root_meta', None):
            json_api_meta = serializer.get_root_meta(resource, many)
            assert isinstance(json_api_meta, dict), 'get_root_meta must return a dict'
            data.update(json_api_meta)
        return data

    @classmethod
    def build_json_resource_obj(cls, fields, resource, resource_instance, resource_name,
                                force_type_resolution=False):
        """
        Builds the resource object (type, id, attributes) and extracts relationships.
        """
        # Determine type from the instance if the underlying model is polymorphic
        if force_type_resolution:
            resource_name = utils.get_resource_type_from_instance(resource_instance)
        resource_data = [
            ('type', resource_name),
            ('id', encoding.force_str(resource_instance.pk) if resource_instance else None),
            ('attributes', cls.extract_attributes(fields, resource)),
        ]
        relationships = cls.extract_relationships(fields, resource, resource_instance)
        if relationships:
            resource_data.append(('relationships', relationships))
        # Add 'self' link if field is present and valid
        if api_settings.URL_FIELD_NAME in resource and \
                isinstance(fields[api_settings.URL_FIELD_NAME], relations.RelatedField):
            resource_data.append(('links', {'self': resource[api_settings.URL_FIELD_NAME]}))
        return OrderedDict(resource_data)

    def render_relationship_view(self, data, accepted_media_type=None, renderer_context=None):
        # Special case for RelationshipView
        view = renderer_context.get("view", None)
        render_data = OrderedDict([
            ('data', data)
        ])
        links = view.get_links()
        if links:
            render_data.update({'links': links}),
        return super(JSONRenderer, self).render(
            render_data, accepted_media_type, renderer_context
        )

    def render_errors(self, data, accepted_media_type=None, renderer_context=None):
        return super(JSONRenderer, self).render(
            utils.format_errors(data), accepted_media_type, renderer_context
        )

    def render(self, data, accepted_media_type=None, renderer_context=None):

        renderer_context = renderer_context or {}

        view = renderer_context.get("view", None)
        request = renderer_context.get("request", None)

        # Get the resource name.
        resource_name = utils.get_resource_name(renderer_context)

        # If this is an error response, skip the rest.
        if resource_name == 'errors':
            return self.render_errors(data, accepted_media_type, renderer_context)

        # if response.status_code is 204 then the data to be rendered must
        # be None
        response = renderer_context.get('response', None)
        if response is not None and response.status_code == 204:
            return super(JSONRenderer, self).render(
                None, accepted_media_type, renderer_context
            )

        from rest_framework_json_api.views import RelationshipView
        if isinstance(view, RelationshipView):
            return self.render_relationship_view(data, accepted_media_type, renderer_context)

        # If `resource_name` is set to None then render default as the dev
        # wants to build the output format manually.
        if resource_name is None or resource_name is False:
            return super(JSONRenderer, self).render(
                data, accepted_media_type, renderer_context
            )

        json_api_data = data
        # initialize json_api_meta with pagination meta or an empty dict
        json_api_meta = data.get('meta', {}) if isinstance(data, dict) else {}
        included_cache = defaultdict(dict)

        if data and 'results' in data:
            serializer_data = data["results"]
        else:
            serializer_data = data

        serializer = getattr(serializer_data, 'serializer', None)

        included_resources = utils.get_included_resources(request, serializer)

        if serializer is not None:

            # Extract root meta for any type of serializer
            json_api_meta.update(self.extract_root_meta(serializer, serializer_data))

            if getattr(serializer, 'many', False):
                json_api_data = list()

                for position in range(len(serializer_data)):
                    resource = serializer_data[position]  # Get current resource
                    resource_instance = serializer.instance[position]  # Get current instance

                    if isinstance(serializer.child, rest_framework_json_api.
                                  serializers.PolymorphicModelSerializer):
                        resource_serializer_class = serializer.child.\
                            get_polymorphic_serializer_for_instance(resource_instance)(
                                context=serializer.child.context
                            )
                    else:
                        resource_serializer_class = serializer.child

                    fields = utils.get_serializer_fields(resource_serializer_class)
                    force_type_resolution = getattr(
                        resource_serializer_class, '_poly_force_type_resolution', False)

                    json_resource_obj = self.build_json_resource_obj(
                        fields, resource, resource_instance, resource_name, force_type_resolution
                    )
                    meta = self.extract_meta(serializer, resource)
                    if meta:
                        json_resource_obj.update({'meta': utils.format_field_names(meta)})
                    json_api_data.append(json_resource_obj)

                    self.extract_included(
                        fields, resource, resource_instance, included_resources, included_cache
                    )
            else:
                fields = utils.get_serializer_fields(serializer)
                force_type_resolution = getattr(serializer, '_poly_force_type_resolution', False)

                resource_instance = serializer.instance
                json_api_data = self.build_json_resource_obj(
                    fields, serializer_data, resource_instance, resource_name, force_type_resolution
                )

                meta = self.extract_meta(serializer, serializer_data)
                if meta:
                    json_api_data.update({'meta': utils.format_field_names(meta)})

                self.extract_included(
                    fields, serializer_data, resource_instance, included_resources, included_cache
                )

        # Make sure we render data in a specific order
        render_data = OrderedDict()

        if isinstance(data, dict) and data.get('links'):
            render_data['links'] = data.get('links')

        # format the api root link list
        if view.__class__ and view.__class__.__name__ == 'APIRoot':
            render_data['data'] = None
            render_data['links'] = json_api_data
        else:
            render_data['data'] = json_api_data

        if included_cache:
            if isinstance(json_api_data, list):
                objects = json_api_data
            else:
                objects = [json_api_data]

            for object in objects:
                obj_type = object.get('type')
                obj_id = object.get('id')
                if obj_type in included_cache and \
                   obj_id in included_cache[obj_type]:
                    del included_cache[obj_type][obj_id]
                if not included_cache[obj_type]:
                    del included_cache[obj_type]

        if included_cache:
            render_data['included'] = list()
            for included_type in sorted(included_cache.keys()):
                for included_id in sorted(included_cache[included_type].keys()):
                    render_data['included'].append(included_cache[included_type][included_id])

        if json_api_meta:
            render_data['meta'] = utils.format_field_names(json_api_meta)

        return super(JSONRenderer, self).render(
            render_data, accepted_media_type, renderer_context
        )



from copy import deepcopy

import pytest
from django.urls import reverse
from rest_framework import status

from example import models, serializers, views

pytestmark = pytest.mark.django_db


class _PatchedModel:
    class JSONAPIMeta:
        resource_name = "resource_name_from_JSONAPIMeta"


def _check_resource_and_relationship_comment_type_match(django_client):
    entry_response = django_client.get(reverse("entry-list"))
    comment_response = django_client.get(reverse("comment-list"))

    comment_resource_type = comment_response.json().get('data')[0].get('type')
    comment_relationship_type = entry_response.json().get(
        'data')[0].get('relationships').get('comments').get('data')[0].get('type')

    assert comment_resource_type == comment_relationship_type, (
        "The resource type seen in the relationships and head resource do not match"
    )


def _check_relationship_and_included_comment_type_are_the_same(django_client, url):
    response = django_client.get(url + "?include=comments")
    data = response.json().get('data')[0]
    comment = response.json().get('included')[0]

    comment_relationship_type = data.get('relationships').get('comments').get('data')[0].get('type')
    comment_included_type = comment.get('type')

    assert comment_relationship_type == comment_included_type, (
        "The resource type seen in the relationships and included do not match"
    )


@pytest.mark.usefixtures("single_entry")
class TestModelResourceName:

    create_data = {
        'data': {
            'type': 'resource_name_from_JSONAPIMeta',
            'id': None,
            'attributes': {
                'body': 'example',
            },
            'relationships': {
                'entry': {
                    'data': {
                        'type': 'resource_name_from_JSONAPIMeta',
                        'id': 1
                    }
                }
            }
        }
    }

    def test_model_resource_name_on_list(self, client):
        models.Comment.__bases__ += (_PatchedModel,)
        response = client.get(reverse("comment-list"))
        data = response.json()['data'][0]
        # name should be super-author instead of model name RenamedAuthor
        assert (data.get('type') == 'resource_name_from_JSONAPIMeta'), (
            'resource_name from model incorrect on list')

    # Precedence tests
    def test_resource_name_precendence(self, client, monkeypatch):
        # default
        response = client.get(reverse("comment-list"))
        data = response.json()['data'][0]
        assert (data.get('type') == 'comments'), (
            'resource_name from model incorrect on list')

        # model > default
        models.Comment.__bases__ += (_PatchedModel,)
        response = client.get(reverse("comment-list"))
        data = response.json()['data'][0]
        assert (data.get('type') == 'resource_name_from_JSONAPIMeta'), (
            'resource_name from model incorrect on list')

        # serializer > model
        monkeypatch.setattr(
            serializers.CommentSerializer.Meta,
            'resource_name',
            'resource_name_from_serializer',
            False
        )
        response = client.get(reverse("comment-list"))
        data = response.json()['data'][0]
        assert (data.get('type') == 'resource_name_from_serializer'), (
            'resource_name from serializer incorrect on list')

        # view > serializer > model
        monkeypatch.setattr(views.CommentViewSet, 'resource_name', 'resource_name_from_view', False)
        response = client.get(reverse("comment-list"))
        data = response.json()['data'][0]
        assert (data.get('type') == 'resource_name_from_view'), (
            'resource_name from view incorrect on list')

    def test_model_resource_name_create(self, client):
        models.Comment.__bases__ += (_PatchedModel,)
        models.Entry.__bases__ += (_PatchedModel,)
        response = client.post(reverse("comment-list"), self.create_data)

        assert response.status_code == status.HTTP_201_CREATED

    def test_serializer_resource_name_create(self, client, monkeypatch):
        monkeypatch.setattr(
            serializers.CommentSerializer.Meta,
            'resource_name',
            'renamed_comments',
            False
        )
        monkeypatch.setattr(
            serializers.EntrySerializer.Meta,
            'resource_name',
            'renamed_entries',
            False
        )
        create_data = deepcopy(self.create_data)
        create_data['data']['type'] = 'renamed_comments'
        create_data['data']['relationships']['entry']['data']['type'] = 'renamed_entries'

        response = client.post(reverse("comment-list"), create_data)

        assert response.status_code == status.HTTP_201_CREATED

    def teardown_method(self, method):
        models.Comment.__bases__ = (models.Comment.__bases__[0],)
        models.Entry.__bases__ = (models.Entry.__bases__[0],)


@pytest.mark.usefixtures("single_entry")
class TestResourceNameConsistency:

    # Included rename tests
    def test_type_match_on_included_and_inline_base(self, client):
        _check_relationship_and_included_comment_type_are_the_same(client, reverse("entry-list"))

    def test_type_match_on_included_and_inline_with_JSONAPIMeta(self, client):
        models.Comment.__bases__ += (_PatchedModel,)

        _check_relationship_and_included_comment_type_are_the_same(client, reverse("entry-list"))

    def test_type_match_on_included_and_inline_with_serializer_resource_name(self, client):
        serializers.CommentSerializer.Meta.resource_name = "resource_name_from_serializer"

        _check_relationship_and_included_comment_type_are_the_same(client, reverse("entry-list"))

    def test_type_match_on_included_and_inline_without_serializer_resource_name(self, client):
        serializers.CommentSerializer.Meta.resource_name = None

        _check_relationship_and_included_comment_type_are_the_same(client, reverse("entry-list"))

    def test_type_match_on_included_and_inline_with_serializer_resource_name_and_JSONAPIMeta(
            self, client
    ):
        models.Comment.__bases__ += (_PatchedModel,)
        serializers.CommentSerializer.Meta.resource_name = "resource_name_from_serializer"

        _check_relationship_and_included_comment_type_are_the_same(client, reverse("entry-list"))

    # Relation rename tests
    def test_resource_and_relationship_type_match(self, client):
        _check_resource_and_relationship_comment_type_match(client)

    def test_resource_and_relationship_type_match_with_serializer_resource_name(self, client):
        serializers.CommentSerializer.Meta.resource_name = "resource_name_from_serializer"

        _check_resource_and_relationship_comment_type_match(client)

    def test_resource_and_relationship_type_match_with_JSONAPIMeta(self, client):
        models.Comment.__bases__ += (_PatchedModel,)

        _check_resource_and_relationship_comment_type_match(client)

    def test_resource_and_relationship_type_match_with_serializer_resource_name_and_JSONAPIMeta(
            self, client
    ):
        models.Comment.__bases__ += (_PatchedModel,)
        serializers.CommentSerializer.Meta.resource_name = "resource_name_from_serializer"

        _check_resource_and_relationship_comment_type_match(client)

    def teardown_method(self, method):
        models.Comment.__bases__ = (models.Comment.__bases__[0],)
        try:
            delattr(serializers.CommentSerializer.Meta, "resource_name")
        except AttributeError:
            pass



# Generated by Django 3.0.3 on 2020-02-06 10:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('example', '0007_artproject_description'),
    ]

    operations = [
        migrations.CreateModel(
            name='LabResults',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField()),
                ('measurements', models.TextField()),
                ('research_project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='lab_results', to='example.ResearchProject')),
            ],
        ),
    ]



"""
This module provides the `json_api_settings` object that is used to access
JSON API REST framework settings, checking for user settings first, then falling back to
the defaults.
"""

from django.conf import settings
from django.core.signals import setting_changed

JSON_API_SETTINGS_PREFIX = 'JSON_API_'

DEFAULTS = {
    'FORMAT_FIELD_NAMES': False,
    'FORMAT_TYPES': False,
    'PLURALIZE_TYPES': False,
    'UNIFORM_EXCEPTIONS': False,
    'SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE': False
}


class JSONAPISettings(object):
    """
    A settings object that allows json api settings to be access as
    properties.
    """

    def __init__(self, user_settings=settings, defaults=DEFAULTS):
        self.defaults = defaults
        self.user_settings = user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid JSON API setting: '%s'" % attr)

        value = getattr(self.user_settings, JSON_API_SETTINGS_PREFIX + attr, self.defaults[attr])

        # Cache the result
        setattr(self, attr, value)
        return value


json_api_settings = JSONAPISettings()


def reload_json_api_settings(*args, **kwargs):
    django_setting = kwargs['setting']
    setting = django_setting.replace(JSON_API_SETTINGS_PREFIX, '')
    value = kwargs['value']
    if setting in DEFAULTS.keys():
        if value is not None:
            setattr(json_api_settings, setting, value)
        elif hasattr(json_api_settings, setting):
            delattr(json_api_settings, setting)


setting_changed.connect(reload_json_api_settings)



import json
import warnings
from collections import OrderedDict

import inflection
from django.core.exceptions import ImproperlyConfigured
from django.urls import NoReverseMatch
from django.utils.translation import gettext_lazy as _
from rest_framework.fields import MISSING_ERROR_MESSAGE, Field, SkipField
from rest_framework.relations import MANY_RELATION_KWARGS
from rest_framework.relations import ManyRelatedField as DRFManyRelatedField
from rest_framework.relations import PrimaryKeyRelatedField, RelatedField
from rest_framework.reverse import reverse
from rest_framework.serializers import Serializer

from rest_framework_json_api.exceptions import Conflict
from rest_framework_json_api.utils import (
    Hyperlink,
    get_included_serializers,
    get_resource_type_from_instance,
    get_resource_type_from_queryset,
    get_resource_type_from_serializer
)

LINKS_PARAMS = [
    'self_link_view_name',
    'related_link_view_name',
    'related_link_lookup_field',
    'related_link_url_kwarg'
]


class SkipDataMixin(object):
    """
    This workaround skips "data" rendering for relationships
    in order to save some sql queries and improve performance
    """

    def __init__(self, *args, **kwargs):
        super(SkipDataMixin, self).__init__(*args, **kwargs)

    def get_attribute(self, instance):
        raise SkipField

    def to_representation(self, *args):
        raise NotImplementedError


class ManyRelatedFieldWithNoData(SkipDataMixin, DRFManyRelatedField):
    pass


class HyperlinkedMixin(object):
    self_link_view_name = None
    related_link_view_name = None
    related_link_lookup_field = 'pk'

    def __init__(self, self_link_view_name=None, related_link_view_name=None, **kwargs):
        if self_link_view_name is not None:
            self.self_link_view_name = self_link_view_name
        if related_link_view_name is not None:
            self.related_link_view_name = related_link_view_name

        self.related_link_lookup_field = kwargs.pop(
            'related_link_lookup_field', self.related_link_lookup_field
        )
        self.related_link_url_kwarg = kwargs.pop(
            'related_link_url_kwarg', self.related_link_lookup_field
        )

        # We include this simply for dependency injection in tests.
        # We can't add it as a class attributes or it would expect an
        # implicit `self` argument to be passed.
        self.reverse = reverse

        super(HyperlinkedMixin, self).__init__(**kwargs)

    def get_url(self, name, view_name, kwargs, request):
        """
        Given a name, view name and kwargs, return the URL that hyperlinks to the object.

        May raise a `NoReverseMatch` if the `view_name` and `lookup_field`
        attributes are not configured to correctly match the URL conf.
        """

        # Return None if the view name is not supplied
        if not view_name:
            return None

        # Return the hyperlink, or error if incorrectly configured.
        try:
            url = self.reverse(view_name, kwargs=kwargs, request=request)
        except NoReverseMatch:
            msg = (
                'Could not resolve URL for hyperlinked relationship using '
                'view name "%s".'
            )
            raise ImproperlyConfigured(msg % view_name)

        if url is None:
            return None

        return Hyperlink(url, name)

    def get_links(self, obj=None, lookup_field='pk'):
        request = self.context.get('request', None)
        view = self.context.get('view', None)
        return_data = OrderedDict()

        kwargs = {lookup_field: getattr(obj, lookup_field) if obj else view.kwargs[lookup_field]}

        self_kwargs = kwargs.copy()
        self_kwargs.update({
            'related_field': self.field_name if self.field_name else self.parent.field_name
        })
        self_link = self.get_url('self', self.self_link_view_name, self_kwargs, request)

        # Assuming RelatedField will be declared in two ways:
        # 1. url(r'^authors/(?P<pk>[^/.]+)/(?P<related_field>\w+)/$',
        #         AuthorViewSet.as_view({'get': 'retrieve_related'}))
        # 2. url(r'^authors/(?P<author_pk>[^/.]+)/bio/$',
        #         AuthorBioViewSet.as_view({'get': 'retrieve'}))
        # So, if related_link_url_kwarg == 'pk' it will add 'related_field' parameter to reverse()
        if self.related_link_url_kwarg == 'pk':
            related_kwargs = self_kwargs
        else:
            related_kwargs = {self.related_link_url_kwarg: kwargs[self.related_link_lookup_field]}

        related_link = self.get_url('related', self.related_link_view_name, related_kwargs, request)

        if self_link:
            return_data.update({'self': self_link})
        if related_link:
            return_data.update({'related': related_link})
        return return_data


class HyperlinkedRelatedField(HyperlinkedMixin, SkipDataMixin, RelatedField):

    @classmethod
    def many_init(cls, *args, **kwargs):
        """
        This method handles creating a parent `ManyRelatedField` instance
        when the `many=True` keyword argument is passed.

        Typically you won't need to override this method.

        Note that we're over-cautious in passing most arguments to both parent
        and child classes in order to try to cover the general case. If you're
        overriding this method you'll probably want something much simpler, eg:

        .. code:: python

            @classmethod
            def many_init(cls, *args, **kwargs):
                kwargs['child'] = cls()
                return CustomManyRelatedField(*args, **kwargs)
        """
        list_kwargs = {'child_relation': cls(*args, **kwargs)}
        for key in kwargs:
            if key in MANY_RELATION_KWARGS:
                list_kwargs[key] = kwargs[key]
        return ManyRelatedFieldWithNoData(**list_kwargs)


class ResourceRelatedField(HyperlinkedMixin, PrimaryKeyRelatedField):
    _skip_polymorphic_optimization = True
    self_link_view_name = None
    related_link_view_name = None
    related_link_lookup_field = 'pk'

    default_error_messages = {
        'required': _('This field is required.'),
        'does_not_exist': _('Invalid pk "{pk_value}" - object does not exist.'),
        'incorrect_type': _(
            'Incorrect type. Expected resource identifier object, received {data_type}.'
        ),
        'incorrect_relation_type': _(
            'Incorrect relation type. Expected {relation_type}, received {received_type}.'
        ),
        'missing_type': _('Invalid resource identifier object: missing \'type\' attribute'),
        'missing_id': _('Invalid resource identifier object: missing \'id\' attribute'),
        'no_match': _('Invalid hyperlink - No URL match.'),
    }

    def __init__(self, **kwargs):
        # check for a model class that was passed in for the relation type
        model = kwargs.pop('model', None)
        if model:
            self.model = model

        super(ResourceRelatedField, self).__init__(**kwargs)

    def use_pk_only_optimization(self):
        # We need the real object to determine its type...
        return self.get_resource_type_from_included_serializer() is not None

    def conflict(self, key, **kwargs):
        """
        A helper method that simply raises a validation error.
        """
        try:
            msg = self.error_messages[key]
        except KeyError:
            class_name = self.__class__.__name__
            msg = MISSING_ERROR_MESSAGE.format(class_name=class_name, key=key)
            raise AssertionError(msg)
        message_string = msg.format(**kwargs)
        raise Conflict(message_string)

    def to_internal_value(self, data):
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except ValueError:
                # show a useful error if they send a `pk` instead of resource object
                self.fail('incorrect_type', data_type=type(data).__name__)
        if not isinstance(data, dict):
            self.fail('incorrect_type', data_type=type(data).__name__)

        expected_relation_type = get_resource_type_from_queryset(self.get_queryset())
        serializer_resource_type = self.get_resource_type_from_included_serializer()

        if serializer_resource_type is not None:
            expected_relation_type = serializer_resource_type

        if 'type' not in data:
            self.fail('missing_type')

        if 'id' not in data:
            self.fail('missing_id')

        if data['type'] != expected_relation_type:
            self.conflict(
                'incorrect_relation_type',
                relation_type=expected_relation_type,
                received_type=data['type']
            )

        return super(ResourceRelatedField, self).to_internal_value(data['id'])

    def to_representation(self, value):
        if getattr(self, 'pk_field', None) is not None:
            pk = self.pk_field.to_representation(value.pk)
        else:
            pk = value.pk

        resource_type = self.get_resource_type_from_included_serializer()
        if resource_type is None or not self._skip_polymorphic_optimization:
            resource_type = get_resource_type_from_instance(value)

        return OrderedDict([('type', resource_type), ('id', str(pk))])

    def get_resource_type_from_included_serializer(self):
        """
        Check to see it this resource has a different resource_name when
        included and return that name, or None
        """
        field_name = self.field_name or self.parent.field_name
        parent = self.get_parent_serializer()

        if parent is not None:
            # accept both singular and plural versions of field_name
            field_names = [
                inflection.singularize(field_name),
                inflection.pluralize(field_name)
            ]
            includes = get_included_serializers(parent)
            for field in field_names:
                if field in includes.keys():
                    return get_resource_type_from_serializer(includes[field])

        return None

    def get_parent_serializer(self):
        if hasattr(self.parent, 'parent') and self.is_serializer(self.parent.parent):
            return self.parent.parent
        elif self.is_serializer(self.parent):
            return self.parent

        return None

    def is_serializer(self, candidate):
        return isinstance(candidate, Serializer)

    def get_choices(self, cutoff=None):
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field.
            return {}

        if cutoff is not None:
            queryset = queryset[:cutoff]

        return OrderedDict([
            (
                json.dumps(self.to_representation(item)),
                self.display_value(item)
            )
            for item in queryset
        ])


class PolymorphicResourceRelatedField(ResourceRelatedField):
    """
    Inform DRF that the relation must be considered polymorphic.
    Takes a `polymorphic_serializer` as the first positional argument to
    retrieve then validate the accepted types set.
    """

    _skip_polymorphic_optimization = False
    default_error_messages = dict(ResourceRelatedField.default_error_messages, **{
        'incorrect_relation_type': _('Incorrect relation type. Expected one of [{relation_type}], '
                                     'received {received_type}.'),
    })

    def __init__(self, polymorphic_serializer, *args, **kwargs):
        self.polymorphic_serializer = polymorphic_serializer
        super(PolymorphicResourceRelatedField, self).__init__(*args, **kwargs)

    def use_pk_only_optimization(self):
        return False

    def to_internal_value(self, data):
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except ValueError:
                # show a useful error if they send a `pk` instead of resource object
                self.fail('incorrect_type', data_type=type(data).__name__)
        if not isinstance(data, dict):
            self.fail('incorrect_type', data_type=type(data).__name__)

        if 'type' not in data:
            self.fail('missing_type')

        if 'id' not in data:
            self.fail('missing_id')

        expected_relation_types = self.polymorphic_serializer.get_polymorphic_types()

        if data['type'] not in expected_relation_types:
            self.conflict('incorrect_relation_type', relation_type=", ".join(
                expected_relation_types), received_type=data['type'])

        return super(ResourceRelatedField, self).to_internal_value(data['id'])


class SerializerMethodFieldBase(Field):
    def __init__(self, method_name=None, **kwargs):
        if not method_name and kwargs.get('source'):
            method_name = kwargs.pop('source')
            warnings.warn(DeprecationWarning(
                "'source' argument of {cls} is deprecated, use 'method_name' "
                "as in SerializerMethodField".format(cls=self.__class__.__name__)), stacklevel=3)
        self.method_name = method_name
        kwargs['source'] = '*'
        kwargs['read_only'] = True
        super().__init__(**kwargs)

    def bind(self, field_name, parent):
        default_method_name = 'get_{field_name}'.format(field_name=field_name)
        if self.method_name is None:
            self.method_name = default_method_name
        super().bind(field_name, parent)

    def get_attribute(self, instance):
        serializer_method = getattr(self.parent, self.method_name)
        return serializer_method(instance)


class ManySerializerMethodResourceRelatedField(SerializerMethodFieldBase, ResourceRelatedField):
    def __init__(self, child_relation=None, *args, **kwargs):
        assert child_relation is not None, '`child_relation` is a required argument.'
        self.child_relation = child_relation
        super().__init__(**kwargs)
        self.child_relation.bind(field_name='', parent=self)

    def to_representation(self, value):
        return [self.child_relation.to_representation(item) for item in value]


class SerializerMethodResourceRelatedField(SerializerMethodFieldBase, ResourceRelatedField):
    """
    Allows us to use serializer method RelatedFields
    with return querysets
    """

    many_kwargs = [*MANY_RELATION_KWARGS, *LINKS_PARAMS, 'method_name', 'model']
    many_cls = ManySerializerMethodResourceRelatedField

    @classmethod
    def many_init(cls, *args, **kwargs):
        list_kwargs = {'child_relation': cls(**kwargs)}
        for key in kwargs:
            if key in cls.many_kwargs:
                list_kwargs[key] = kwargs[key]
        return cls.many_cls(**list_kwargs)


class ManySerializerMethodHyperlinkedRelatedField(SkipDataMixin,
                                                  ManySerializerMethodResourceRelatedField):
    pass


class SerializerMethodHyperlinkedRelatedField(SkipDataMixin,
                                              SerializerMethodResourceRelatedField):
    many_cls = ManySerializerMethodHyperlinkedRelatedField



# Generated by Django 2.2.2 on 2019-06-07 06:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('example', '0006_auto_20181228_0752'),
    ]

    operations = [
        migrations.AddField(
            model_name='artproject',
            name='description',
            field=models.CharField(max_length=100, null=True),
        ),
    ]



"""
Parsers
"""
from rest_framework import parsers
from rest_framework.exceptions import ParseError

from . import exceptions, renderers, serializers, utils
from .settings import json_api_settings


class JSONParser(parsers.JSONParser):
    """
    Similar to `JSONRenderer`, the `JSONParser` you may override the following methods if you
    need highly custom parsing control.

    A JSON API client will send a payload that looks like this:

    .. code:: json


        {
            "data": {
                "type": "identities",
                "id": 1,
                "attributes": {
                    "first_name": "John",
                    "last_name": "Coltrane"
                }
            }
        }

    We extract the attributes so that DRF serializers can work as normal.
    """
    media_type = 'application/vnd.api+json'
    renderer_class = renderers.JSONRenderer

    @staticmethod
    def parse_attributes(data):
        attributes = data.get('attributes')
        uses_format_translation = json_api_settings.FORMAT_FIELD_NAMES

        if not attributes:
            return dict()
        elif uses_format_translation:
            # convert back to python/rest_framework's preferred underscore format
            return utils.format_field_names(attributes, 'underscore')
        else:
            return attributes

    @staticmethod
    def parse_relationships(data):
        uses_format_translation = json_api_settings.FORMAT_FIELD_NAMES
        relationships = data.get('relationships')

        if not relationships:
            relationships = dict()
        elif uses_format_translation:
            # convert back to python/rest_framework's preferred underscore format
            relationships = utils.format_field_names(relationships, 'underscore')

        # Parse the relationships
        parsed_relationships = dict()
        for field_name, field_data in relationships.items():
            field_data = field_data.get('data')
            if isinstance(field_data, dict) or field_data is None:
                parsed_relationships[field_name] = field_data
            elif isinstance(field_data, list):
                parsed_relationships[field_name] = list(relation for relation in field_data)
        return parsed_relationships

    @staticmethod
    def parse_metadata(result):
        """
        Returns a dictionary which will be merged into parsed data of the request. By default,
        it reads the `meta` content in the request body and returns it in a dictionary with
        a `_meta` top level key.
        """
        metadata = result.get('meta')
        if metadata:
            return {'_meta': metadata}
        else:
            return {}

    def parse(self, stream, media_type=None, parser_context=None):
        """
        Parses the incoming bytestream as JSON and returns the resulting data
        """
        result = super(JSONParser, self).parse(
            stream, media_type=media_type, parser_context=parser_context
        )

        if not isinstance(result, dict) or 'data' not in result:
            raise ParseError('Received document does not contain primary data')

        data = result.get('data')
        view = parser_context['view']

        from rest_framework_json_api.views import RelationshipView
        if isinstance(view, RelationshipView):
            # We skip parsing the object as JSONAPI Resource Identifier Object and not a regular
            # Resource Object
            if isinstance(data, list):
                for resource_identifier_object in data:
                    if not (
                        resource_identifier_object.get('id') and
                        resource_identifier_object.get('type')
                    ):
                        raise ParseError(
                            'Received data contains one or more malformed JSONAPI '
                            'Resource Identifier Object(s)'
                        )
            elif not (data.get('id') and data.get('type')):
                raise ParseError('Received data is not a valid JSONAPI Resource Identifier Object')

            return data

        request = parser_context.get('request')

        # Sanity check
        if not isinstance(data, dict):
            raise ParseError('Received data is not a valid JSONAPI Resource Identifier Object')

        # Check for inconsistencies
        if request.method in ('PUT', 'POST', 'PATCH'):
            resource_name = utils.get_resource_name(
                parser_context, expand_polymorphic_types=True)
            if isinstance(resource_name, str):
                if data.get('type') != resource_name:
                    raise exceptions.Conflict(
                        "The resource object's type ({data_type}) is not the type that "
                        "constitute the collection represented by the endpoint "
                        "({resource_type}).".format(
                            data_type=data.get('type'),
                            resource_type=resource_name))
            else:
                if data.get('type') not in resource_name:
                    raise exceptions.Conflict(
                        "The resource object's type ({data_type}) is not the type that "
                        "constitute the collection represented by the endpoint "
                        "(one of [{resource_types}]).".format(
                            data_type=data.get('type'),
                            resource_types=", ".join(resource_name)))
        if not data.get('id') and request.method in ('PATCH', 'PUT'):
            raise ParseError("The resource identifier object must contain an 'id' member")

        if request.method in ('PATCH', 'PUT'):
            lookup_url_kwarg = getattr(view, 'lookup_url_kwarg', None) or \
                getattr(view, 'lookup_field', None)
            if lookup_url_kwarg and str(data.get('id')) != str(view.kwargs[lookup_url_kwarg]):
                raise exceptions.Conflict(
                    "The resource object's id ({data_id}) does not match url's "
                    "lookup id ({url_id})".format(
                        data_id=data.get('id'),
                        url_id=view.kwargs[lookup_url_kwarg]
                    )
                )

        # Construct the return data
        serializer_class = getattr(view, 'serializer_class', None)
        parsed_data = {'id': data.get('id')} if 'id' in data else {}
        # `type` field needs to be allowed in none polymorphic serializers
        if serializer_class is not None:
            if issubclass(serializer_class, serializers.PolymorphicModelSerializer):
                parsed_data['type'] = data.get('type')
        parsed_data.update(self.parse_attributes(data))
        parsed_data.update(self.parse_relationships(data))
        parsed_data.update(self.parse_metadata(result))
        return parsed_data



from collections import OrderedDict

from rest_framework.request import Request
from rest_framework.test import APIRequestFactory
from rest_framework.utils.urls import replace_query_param

from rest_framework_json_api import pagination

factory = APIRequestFactory()


class TestLimitOffset:
    """
    Unit tests for `pagination.JsonApiLimitOffsetPagination`.
    """

    def setup(self):
        class ExamplePagination(pagination.JsonApiLimitOffsetPagination):
            default_limit = 10
            max_limit = 15

        self.pagination = ExamplePagination()
        self.queryset = range(1, 101)
        self.base_url = 'http://testserver/'

    def paginate_queryset(self, request):
        return list(self.pagination.paginate_queryset(self.queryset, request))

    def get_paginated_content(self, queryset):
        response = self.pagination.get_paginated_response(queryset)
        return response.data

    def get_test_request(self, arguments):
        return Request(factory.get('/', arguments))

    def test_valid_offset_limit(self):
        """
        Basic test, assumes offset and limit are given.
        """
        offset = 10
        limit = 5
        count = len(self.queryset)
        last_offset = (count // limit) * limit
        next_offset = 15
        prev_offset = 5

        request = self.get_test_request({
            self.pagination.limit_query_param: limit,
            self.pagination.offset_query_param: offset
        })
        base_url = replace_query_param(self.base_url, self.pagination.limit_query_param, limit)
        last_url = replace_query_param(base_url, self.pagination.offset_query_param, last_offset)
        first_url = base_url
        next_url = replace_query_param(base_url, self.pagination.offset_query_param, next_offset)
        prev_url = replace_query_param(base_url, self.pagination.offset_query_param, prev_offset)
        queryset = self.paginate_queryset(request)
        content = self.get_paginated_content(queryset)
        next_offset = offset + limit

        expected_content = {
            'results': list(range(offset + 1, next_offset + 1)),
            'links': OrderedDict([
                ('first', first_url),
                ('last', last_url),
                ('next', next_url),
                ('prev', prev_url),
            ]),
            'meta': {
                'pagination': OrderedDict([
                    ('count', count),
                    ('limit', limit),
                    ('offset', offset),
                ])
            }
        }

        assert queryset == list(range(offset + 1, next_offset + 1))
        assert content == expected_content



import os

SITE_ID = 1
DEBUG = True

MEDIA_ROOT = os.path.normcase(os.path.dirname(os.path.abspath(__file__)))
MEDIA_URL = '/media/'

DATABASE_ENGINE = 'sqlite3'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'drf_example',
    }
}

INSTALLED_APPS = [
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    'django.contrib.sessions',
    'django.contrib.auth',
    'rest_framework',
    'polymorphic',
    'example',
    'debug_toolbar',
    'django_filters',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            # insert your TEMPLATE_DIRS here
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                # Insert your TEMPLATE_CONTEXT_PROCESSORS here or use this
                # list if you haven't customized them:
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

STATIC_URL = '/static/'

ROOT_URLCONF = 'example.urls'

SECRET_KEY = 'abc123'

PASSWORD_HASHERS = ('django.contrib.auth.hashers.UnsaltedMD5PasswordHasher', )

MIDDLEWARE = (
    'debug_toolbar.middleware.DebugToolbarMiddleware',
)

INTERNAL_IPS = ('127.0.0.1', )

JSON_API_FORMAT_FIELD_NAMES = 'camelize'
JSON_API_FORMAT_TYPES = 'camelize'
JSON_API_SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE = True
REST_FRAMEWORK = {
    'PAGE_SIZE': 5,
    'EXCEPTION_HANDLER': 'rest_framework_json_api.exceptions.exception_handler',
    'DEFAULT_PAGINATION_CLASS':
        'rest_framework_json_api.pagination.JsonApiPageNumberPagination',
    'DEFAULT_PARSER_CLASSES': (
        'rest_framework_json_api.parsers.JSONParser',
        'rest_framework.parsers.FormParser',
        'rest_framework.parsers.MultiPartParser'
    ),
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework_json_api.renderers.JSONRenderer',

        # If you're performance testing, you will want to use the browseable API
        # without forms, as the forms can generate their own queries.
        # If performance testing, enable:
        # 'example.utils.BrowsableAPIRendererWithoutForms',
        # Otherwise, to play around with the browseable API, enable:
        'rest_framework.renderers.BrowsableAPIRenderer',
    ),
    'DEFAULT_METADATA_CLASS': 'rest_framework_json_api.metadata.JSONAPIMetadata',
    'DEFAULT_FILTER_BACKENDS': (
        'rest_framework_json_api.filters.OrderingFilter',
        'rest_framework_json_api.django_filters.DjangoFilterBackend',
        'rest_framework.filters.SearchFilter',
    ),
    'SEARCH_PARAM': 'filter[search]',
    'TEST_REQUEST_RENDERER_CLASSES': (
        'rest_framework_json_api.renderers.JSONRenderer',
    ),
    'TEST_REQUEST_DEFAULT_FORMAT': 'vnd.api+json'
}



import json
from io import BytesIO

from django.conf.urls import url
from django.test import TestCase, override_settings
from django.urls import reverse
from rest_framework import status, views
from rest_framework.exceptions import ParseError
from rest_framework.response import Response
from rest_framework.test import APITestCase

from rest_framework_json_api import serializers
from rest_framework_json_api.parsers import JSONParser
from rest_framework_json_api.renderers import JSONRenderer


class TestJSONParser(TestCase):

    def setUp(self):
        class MockRequest(object):

            def __init__(self):
                self.method = 'GET'

        request = MockRequest()

        self.parser_context = {'request': request, 'kwargs': {}, 'view': 'BlogViewSet'}

        data = {
            'data': {
                'id': 123,
                'type': 'Blog',
                'attributes': {
                    'json-value': {'JsonKey': 'JsonValue'}
                },
            },
            'meta': {
                'random_key': 'random_value'
            }
        }

        self.string = json.dumps(data)

    @override_settings(JSON_API_FORMAT_FIELD_NAMES='dasherize')
    def test_parse_include_metadata_format_field_names(self):
        parser = JSONParser()

        stream = BytesIO(self.string.encode('utf-8'))
        data = parser.parse(stream, None, self.parser_context)

        self.assertEqual(data['_meta'], {'random_key': 'random_value'})
        self.assertEqual(data['json_value'], {'JsonKey': 'JsonValue'})

    def test_parse_invalid_data(self):
        parser = JSONParser()

        string = json.dumps([])
        stream = BytesIO(string.encode('utf-8'))

        with self.assertRaises(ParseError):
            parser.parse(stream, None, self.parser_context)

    def test_parse_invalid_data_key(self):
        parser = JSONParser()

        string = json.dumps({
            'data': [{
                'id': 123,
                'type': 'Blog',
                'attributes': {
                    'json-value': {'JsonKey': 'JsonValue'}
                },
            }]
        })
        stream = BytesIO(string.encode('utf-8'))

        with self.assertRaises(ParseError):
            parser.parse(stream, None, self.parser_context)


class DummyDTO:
    def __init__(self, response_dict):
        for k, v in response_dict.items():
            setattr(self, k, v)

    @property
    def pk(self):
        return self.id if hasattr(self, 'id') else None


class DummySerializer(serializers.Serializer):
    body = serializers.CharField()
    id = serializers.IntegerField()


class DummyAPIView(views.APIView):
    parser_classes = [JSONParser]
    renderer_classes = [JSONRenderer]
    resource_name = 'dummy'

    def patch(self, request, *args, **kwargs):
        serializer = DummySerializer(DummyDTO(request.data))
        return Response(status=status.HTTP_200_OK, data=serializer.data)


urlpatterns = [
    url(r'repeater$', DummyAPIView.as_view(), name='repeater'),
]


class TestParserOnAPIView(APITestCase):

    def setUp(self):
        class MockRequest(object):
            def __init__(self):
                self.method = 'PATCH'

        request = MockRequest()
        # To be honest view string isn't resolved into actual view
        self.parser_context = {'request': request, 'kwargs': {}, 'view': 'DummyAPIView'}

        self.data = {
            'data': {
                'id': 123,
                'type': 'strs',
                'attributes': {
                    'body': 'hello'
                },
            }
        }

        self.string = json.dumps(self.data)

    def test_patch_doesnt_raise_attribute_error(self):
        parser = JSONParser()

        stream = BytesIO(self.string.encode('utf-8'))

        data = parser.parse(stream, None, self.parser_context)

        assert data['id'] == 123
        assert data['body'] == 'hello'

    @override_settings(ROOT_URLCONF=__name__)
    def test_patch_request(self):
        url = reverse('repeater')
        data = self.data
        data['data']['type'] = 'dummy'
        response = self.client.patch(url, data=data)
        data = response.json()

        assert data['data']['id'] == str(123)
        assert data['data']['attributes']['body'] == 'hello'



import pytest

from example.factories import BlogFactory
from example.models import Blog

pytestmark = pytest.mark.django_db


def test_factory_instance(blog_factory):

    assert blog_factory == BlogFactory


def test_model_instance(blog):

    assert isinstance(blog, Blog)


def test_multiple_blog(blog_factory):
    another_blog = blog_factory(name='Cool Blog')
    new_blog = blog_factory(name='Awesome Blog')

    assert another_blog.name == 'Cool Blog'
    assert new_blog.name == 'Awesome Blog'


def test_factories_with_relations(author_factory, entry_factory):

    author = author_factory(name="Joel Spolsky")
    entry = entry_factory(
        headline=("The Absolute Minimum Every Software Developer"
                  "Absolutely, Positively Must Know About Unicode "
                  "and Character Sets (No Excuses!)"),
        blog__name='Joel on Software', authors=(author, ))

    assert entry.blog.name == 'Joel on Software'
    assert entry.headline == ("The Absolute Minimum Every Software Developer"
                              "Absolutely, Positively Must Know About Unicode "
                              "and Character Sets (No Excuses!)")
    assert entry.authors.all().count() == 1
    assert entry.authors.all()[0].name == 'Joel Spolsky'



# Generated by Django 2.1.1 on 2018-09-22 15:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('example', '0004_auto_20171011_0631'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('modified_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=50)),
            ],
            options={
                'ordering': ('id',),
            },
        ),
        migrations.AlterModelOptions(
            name='artproject',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='project',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='researchproject',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AddField(
            model_name='project',
            name='project_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='example.ProjectType'),
        ),
    ]



from datetime import datetime

from rest_framework import fields as drf_fields
from rest_framework import serializers as drf_serilazers

from rest_framework_json_api import relations, serializers

from example.models import (
    ArtProject,
    Author,
    AuthorBio,
    AuthorBioMetadata,
    AuthorType,
    Blog,
    Comment,
    Company,
    Entry,
    LabResults,
    Project,
    ProjectType,
    ResearchProject,
    TaggedItem
)


class TaggedItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = TaggedItem
        fields = ('tag',)


class TaggedItemDRFSerializer(drf_serilazers.ModelSerializer):
    """
    DRF default serializer to test default DRF functionalities
    """
    class Meta:
        model = TaggedItem
        fields = ('tag',)


class BlogSerializer(serializers.ModelSerializer):
    copyright = serializers.SerializerMethodField()
    tags = relations.ResourceRelatedField(many=True, read_only=True)

    included_serializers = {
        'tags': 'example.serializers.TaggedItemSerializer',
    }

    def get_copyright(self, resource):
        return datetime.now().year

    def get_root_meta(self, resource, many):
        return {
            'api_docs': '/docs/api/blogs'
        }

    class Meta:
        model = Blog
        fields = ('name', 'url', 'tags')
        read_only_fields = ('tags',)
        meta_fields = ('copyright',)


class BlogDRFSerializer(drf_serilazers.ModelSerializer):
    """
    DRF default serializer to test default DRF functionalities
    """
    copyright = serializers.SerializerMethodField()
    tags = TaggedItemDRFSerializer(many=True, read_only=True)

    def get_copyright(self, resource):
        return datetime.now().year

    def get_root_meta(self, resource, many):
        return {
            'api_docs': '/docs/api/blogs'
        }

    class Meta:
        model = Blog
        fields = ('name', 'url', 'tags', 'copyright')
        read_only_fields = ('tags',)
        meta_fields = ('copyright',)


class EntrySerializer(serializers.ModelSerializer):
    def __init__(self, *args, **kwargs):
        super(EntrySerializer, self).__init__(*args, **kwargs)
        # to make testing more concise we'll only output the
        # `featured` field when it's requested via `include`
        request = kwargs.get('context', {}).get('request')
        if request and 'featured' not in request.query_params.get('include', []):
            self.fields.pop('featured', None)

    included_serializers = {
        'authors': 'example.serializers.AuthorSerializer',
        'comments': 'example.serializers.CommentSerializer',
        'featured': 'example.serializers.EntrySerializer',
        'suggested': 'example.serializers.EntrySerializer',
        'tags': 'example.serializers.TaggedItemSerializer',
    }

    body_format = serializers.SerializerMethodField()
    # single related from model
    blog_hyperlinked = relations.HyperlinkedRelatedField(
        related_link_view_name='entry-blog',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        read_only=True,
        source='blog'
    )
    # many related from model
    comments = relations.ResourceRelatedField(
        many=True, read_only=True)
    # many related hyperlinked from model
    comments_hyperlinked = relations.HyperlinkedRelatedField(
        related_link_view_name='entry-comments',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        many=True,
        read_only=True,
        source='comments'
    )
    # many related from serializer
    suggested = relations.SerializerMethodResourceRelatedField(
        related_link_view_name='entry-suggested',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        model=Entry,
        many=True,
    )
    # many related hyperlinked from serializer
    suggested_hyperlinked = relations.SerializerMethodHyperlinkedRelatedField(
        related_link_view_name='entry-suggested',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        model=Entry,
        many=True,
    )
    # single related from serializer
    featured = relations.SerializerMethodResourceRelatedField(model=Entry)
    # single related hyperlinked from serializer
    featured_hyperlinked = relations.SerializerMethodHyperlinkedRelatedField(
        related_link_view_name='entry-featured',
        related_link_url_kwarg='entry_pk',
        self_link_view_name='entry-relationships',
        model=Entry,
        read_only=True
    )
    tags = relations.ResourceRelatedField(many=True, read_only=True)

    def get_suggested(self, obj):
        return Entry.objects.exclude(pk=obj.pk)

    def get_featured(self, obj):
        return Entry.objects.exclude(pk=obj.pk).first()

    def get_body_format(self, obj):
        return 'text'

    class Meta:
        model = Entry
        fields = ('blog', 'blog_hyperlinked', 'headline', 'body_text', 'pub_date', 'mod_date',
                  'authors', 'comments', 'comments_hyperlinked', 'featured', 'suggested',
                  'suggested_hyperlinked', 'tags', 'featured_hyperlinked')
        read_only_fields = ('tags',)
        meta_fields = ('body_format',)

    class JSONAPIMeta:
        included_resources = ['comments']


class EntryDRFSerializers(drf_serilazers.ModelSerializer):

    tags = TaggedItemDRFSerializer(many=True, read_only=True)
    url = drf_serilazers.HyperlinkedIdentityField(
        view_name='drf-entry-blog-detail',
        lookup_url_kwarg='entry_pk',
        read_only=True,
    )

    class Meta:
        model = Entry
        fields = ('tags', 'url',)
        read_only_fields = ('tags',)


class AuthorTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = AuthorType
        fields = ('name', )


class AuthorBioSerializer(serializers.ModelSerializer):
    class Meta:
        model = AuthorBio
        fields = ('author', 'body', 'metadata')

    included_serializers = {
        'metadata': 'example.serializers.AuthorBioMetadataSerializer',
    }


class AuthorBioMetadataSerializer(serializers.ModelSerializer):
    class Meta:
        model = AuthorBioMetadata
        fields = ('body',)


class AuthorSerializer(serializers.ModelSerializer):
    bio = relations.ResourceRelatedField(
        related_link_view_name='author-related',
        self_link_view_name='author-relationships',
        queryset=AuthorBio.objects,
    )
    entries = relations.ResourceRelatedField(
        related_link_view_name='author-related',
        self_link_view_name='author-relationships',
        queryset=Entry.objects,
        many=True
    )
    first_entry = relations.SerializerMethodResourceRelatedField(
        related_link_view_name='author-related',
        self_link_view_name='author-relationships',
        model=Entry,
    )
    comments = relations.HyperlinkedRelatedField(
        related_link_view_name='author-related',
        self_link_view_name='author-relationships',
        queryset=Comment.objects,
        many=True
    )
    included_serializers = {
        'bio': AuthorBioSerializer,
        'type': AuthorTypeSerializer
    }
    related_serializers = {
        'bio': 'example.serializers.AuthorBioSerializer',
        'type': 'example.serializers.AuthorTypeSerializer',
        'comments': 'example.serializers.CommentSerializer',
        'entries': 'example.serializers.EntrySerializer',
        'first_entry': 'example.serializers.EntrySerializer'
    }

    class Meta:
        model = Author
        fields = ('name', 'email', 'bio', 'entries', 'comments', 'first_entry', 'type')

    def get_first_entry(self, obj):
        return obj.entries.first()


class WriterSerializer(serializers.ModelSerializer):
    included_serializers = {
        'bio': AuthorBioSerializer
    }

    class Meta:
        model = Author
        fields = ('name', 'email', 'bio')
        resource_name = 'writers'


class CommentSerializer(serializers.ModelSerializer):
    # testing remapping of related name
    writer = relations.ResourceRelatedField(source='author', read_only=True)

    included_serializers = {
        'entry': EntrySerializer,
        'author': AuthorSerializer,
        'writer': WriterSerializer
    }

    class Meta:
        model = Comment
        exclude = ('created_at', 'modified_at',)
        # fields = ('entry', 'body', 'author',)


class ProjectTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = ProjectType
        fields = ('name', 'url',)


class BaseProjectSerializer(serializers.ModelSerializer):
    included_serializers = {
        'project_type': ProjectTypeSerializer,
    }


class ArtProjectSerializer(BaseProjectSerializer):
    class Meta:
        model = ArtProject
        exclude = ('polymorphic_ctype',)


class ResearchProjectSerializer(BaseProjectSerializer):
    # testing exclusive related field on inherited polymorphic model
    lab_results = relations.ResourceRelatedField(many=True, read_only=True)

    class Meta:
        model = ResearchProject
        exclude = ('polymorphic_ctype',)


class LabResultsSerializer(serializers.ModelSerializer):
    class Meta:
        model = LabResults
        fields = ('date', 'measurements')


class ProjectSerializer(serializers.PolymorphicModelSerializer):
    included_serializers = {
        'project_type': ProjectTypeSerializer,
    }
    polymorphic_serializers = [ArtProjectSerializer, ResearchProjectSerializer]

    class Meta:
        model = Project
        exclude = ('polymorphic_ctype',)


class CurrentProjectRelatedField(relations.PolymorphicResourceRelatedField):
    def get_attribute(self, instance):
        obj = super(CurrentProjectRelatedField, self).get_attribute(instance)

        is_art = (
            self.field_name == 'current_art_project' and
            isinstance(obj, ArtProject)
        )
        is_res = (
            self.field_name == 'current_research_project' and
            isinstance(obj, ResearchProject)
        )

        if is_art or is_res:
            return obj

        raise drf_fields.SkipField()


class CompanySerializer(serializers.ModelSerializer):
    current_project = relations.PolymorphicResourceRelatedField(
        ProjectSerializer, queryset=Project.objects.all())
    current_art_project = CurrentProjectRelatedField(
        ProjectSerializer, source='current_project', read_only=True)
    current_research_project = CurrentProjectRelatedField(
        ProjectSerializer, source='current_project', read_only=True)
    future_projects = relations.PolymorphicResourceRelatedField(
        ProjectSerializer, queryset=Project.objects.all(), many=True)

    included_serializers = {
        'current_project': ProjectSerializer,
        'future_projects': ProjectSerializer,
        'current_art_project': ProjectSerializer,
        'current_research_project': ProjectSerializer
    }

    class Meta:
        model = Company
        fields = '__all__'



from django.conf import settings
from django.conf.urls import include, url
from rest_framework import routers

from example.views import (
    AuthorRelationshipView,
    AuthorViewSet,
    BlogRelationshipView,
    BlogViewSet,
    CommentRelationshipView,
    CommentViewSet,
    CompanyViewset,
    EntryRelationshipView,
    EntryViewSet,
    NonPaginatedEntryViewSet,
    ProjectTypeViewset,
    ProjectViewset
)

router = routers.DefaultRouter(trailing_slash=False)

router.register(r'blogs', BlogViewSet)
router.register(r'entries', EntryViewSet)
router.register(r'nopage-entries', NonPaginatedEntryViewSet, 'nopage-entry')
router.register(r'authors', AuthorViewSet)
router.register(r'comments', CommentViewSet)
router.register(r'companies', CompanyViewset)
router.register(r'projects', ProjectViewset)
router.register(r'project-types', ProjectTypeViewset)

urlpatterns = [
    url(r'^', include(router.urls)),
    url(r'^entries/(?P<entry_pk>[^/.]+)/suggested/$',
        EntryViewSet.as_view({'get': 'list'}),
        name='entry-suggested'
        ),
    url(r'entries/(?P<entry_pk>[^/.]+)/blog$',
        BlogViewSet.as_view({'get': 'retrieve'}),
        name='entry-blog'),
    url(r'entries/(?P<entry_pk>[^/.]+)/comments$',
        CommentViewSet.as_view({'get': 'list'}),
        name='entry-comments'),
    url(r'entries/(?P<entry_pk>[^/.]+)/authors$',
        AuthorViewSet.as_view({'get': 'list'}),
        name='entry-authors'),
    url(r'entries/(?P<entry_pk>[^/.]+)/featured$',
        EntryViewSet.as_view({'get': 'retrieve'}),
        name='entry-featured'),

    url(r'^authors/(?P<pk>[^/.]+)/(?P<related_field>\w+)/$',
        AuthorViewSet.as_view({'get': 'retrieve_related'}),
        name='author-related'),

    url(r'^entries/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$',
        EntryRelationshipView.as_view(),
        name='entry-relationships'),
    url(r'^blogs/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$',
        BlogRelationshipView.as_view(),
        name='blog-relationships'),
    url(r'^comments/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$',
        CommentRelationshipView.as_view(),
        name='comment-relationships'),
    url(r'^authors/(?P<pk>[^/.]+)/relationships/(?P<related_field>\w+)$',
        AuthorRelationshipView.as_view(),
        name='author-relationships'),
]


if settings.DEBUG:
    import debug_toolbar
    urlpatterns = [
        url(r'^__debug__/', include(debug_toolbar.urls)),
    ] + urlpatterns



"""
Test sideloading resources
"""
import json

from django.urls import reverse
from django.utils import encoding

from example.tests import TestBase


class SideloadResourceTest(TestBase):
    """
    Test that sideloading resources returns expected output.
    """
    url = reverse('user-posts')

    def test_get_sideloaded_data(self):
        """
        Ensure resources that are meant for sideloaded data
        do not return a single root key.
        """
        response = self.client.get(self.url)
        content = json.loads(response.content.decode('utf8'))

        self.assertEqual(
            sorted(content.keys()),
            [encoding.force_str('identities'),
             encoding.force_str('posts')])



#!/usr/bin/env python
from __future__ import print_function

import os
import re
import sys

from setuptools import setup

needs_wheel = {'bdist_wheel'}.intersection(sys.argv)
wheel = ['wheel'] if needs_wheel else []


def read(*paths):
    """
    Build a file path from paths and return the contents.
    """
    with open(os.path.join(*paths), 'r') as f:
        return f.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    print("You probably want to also tag the version now:")
    print("  git tag -a {0} -m 'version {0}'".format(
        get_version('rest_framework_json_api')))
    print("  git push --tags")
    sys.exit()

setup(
    name='djangorestframework-jsonapi',
    version=get_version('rest_framework_json_api'),
    url='https://github.com/django-json-api/django-rest-framework-json-api',
    license='BSD',
    description='A Django REST framework API adapter for the JSON API spec.',
    long_description=read('README.rst'),
    author='Jerel Unruh',
    author_email='',
    packages=get_packages('rest_framework_json_api'),
    package_data=get_package_data('rest_framework_json_api'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'inflection>=0.3.0',
        'djangorestframework>=3.10,<3.12',
        'django>=1.11,<3.1',
    ],
    extras_require={
        'django-polymorphic': ['django-polymorphic>=2.0'],
        'django-filter': ['django-filter>=2.0']
    },
    setup_requires=wheel,
    python_requires=">=3.5",
    zip_safe=False,
)



import warnings

import inflection
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.query import QuerySet
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import ParseError
from rest_framework.serializers import *  # noqa: F403

from rest_framework_json_api.exceptions import Conflict
from rest_framework_json_api.relations import ResourceRelatedField
from rest_framework_json_api.settings import json_api_settings
from rest_framework_json_api.utils import (
    get_included_resources,
    get_included_serializers,
    get_resource_type_from_instance,
    get_resource_type_from_model,
    get_resource_type_from_serializer
)


class ResourceIdentifierObjectSerializer(BaseSerializer):
    default_error_messages = {
        'incorrect_model_type': _(
            'Incorrect model type. Expected {model_type}, received {received_type}.'
        ),
        'does_not_exist': _('Invalid pk "{pk_value}" - object does not exist.'),
        'incorrect_type': _('Incorrect type. Expected pk value, received {data_type}.'),
    }

    model_class = None

    def __init__(self, *args, **kwargs):
        self.model_class = kwargs.pop('model_class', self.model_class)
        # this has no fields but assumptions are made elsewhere that self.fields exists.
        self.fields = {}
        super(ResourceIdentifierObjectSerializer, self).__init__(*args, **kwargs)

    def to_representation(self, instance):
        return {
            'type': get_resource_type_from_instance(instance),
            'id': str(instance.pk)
        }

    def to_internal_value(self, data):
        if data['type'] != get_resource_type_from_model(self.model_class):
            self.fail(
                'incorrect_model_type', model_type=self.model_class, received_type=data['type']
            )
        pk = data['id']
        try:
            return self.model_class.objects.get(pk=pk)
        except ObjectDoesNotExist:
            self.fail('does_not_exist', pk_value=pk)
        except (TypeError, ValueError):
            self.fail('incorrect_type', data_type=type(data['pk']).__name__)


class SparseFieldsetsMixin(object):
    def __init__(self, *args, **kwargs):
        super(SparseFieldsetsMixin, self).__init__(*args, **kwargs)
        context = kwargs.get('context')
        request = context.get('request') if context else None

        if request:
            sparse_fieldset_query_param = 'fields[{}]'.format(
                get_resource_type_from_serializer(self)
            )
            try:
                param_name = next(
                    key for key in request.query_params if sparse_fieldset_query_param in key
                )
            except StopIteration:
                pass
            else:
                fieldset = request.query_params.get(param_name).split(',')
                # iterate over a *copy* of self.fields' underlying OrderedDict, because we may
                # modify the original during the iteration.
                # self.fields is a `rest_framework.utils.serializer_helpers.BindingDict`
                for field_name, field in self.fields.fields.copy().items():
                    if field_name == api_settings.URL_FIELD_NAME:  # leave self link there
                        continue
                    if field_name not in fieldset:
                        self.fields.pop(field_name)


class IncludedResourcesValidationMixin(object):
    def __init__(self, *args, **kwargs):
        context = kwargs.get('context')
        request = context.get('request') if context else None
        view = context.get('view') if context else None

        def validate_path(serializer_class, field_path, path):
            serializers = get_included_serializers(serializer_class)
            if serializers is None:
                raise ParseError('This endpoint does not support the include parameter')
            this_field_name = inflection.underscore(field_path[0])
            this_included_serializer = serializers.get(this_field_name)
            if this_included_serializer is None:
                raise ParseError(
                    'This endpoint does not support the include parameter for path {}'.format(
                        path
                    )
                )
            if len(field_path) > 1:
                new_included_field_path = field_path[1:]
                # We go down one level in the path
                validate_path(this_included_serializer, new_included_field_path, path)

        if request and view:
            included_resources = get_included_resources(request)
            for included_field_name in included_resources:
                included_field_path = included_field_name.split('.')
                this_serializer_class = view.get_serializer_class()
                # lets validate the current path
                validate_path(this_serializer_class, included_field_path, included_field_name)

        super(IncludedResourcesValidationMixin, self).__init__(*args, **kwargs)


class SerializerMetaclass(SerializerMetaclass):

    @classmethod
    def _get_declared_fields(cls, bases, attrs):
        fields = super()._get_declared_fields(bases, attrs)
        for field_name, field in fields.items():
            if isinstance(field, BaseSerializer) and \
                    not json_api_settings.SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE:
                clazz = '{}.{}'.format(attrs['__module__'], attrs['__qualname__'])
                if isinstance(field, ListSerializer):
                    nested_class = type(field.child).__name__
                else:
                    nested_class = type(field).__name__

                warnings.warn(DeprecationWarning(
                    "Rendering nested serializer as relationship is deprecated. "
                    "Use `ResourceRelatedField` instead if {} in serializer {} should remain "
                    "a relationship. Otherwise set "
                    "JSON_API_SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE to True to render nested "
                    "serializer as nested json attribute".format(nested_class, clazz)))
        return fields


# If user imports serializer from here we can catch class definition and check
# nested serializers for depricated use.
class Serializer(
    IncludedResourcesValidationMixin, SparseFieldsetsMixin, Serializer,
    metaclass=SerializerMetaclass
):
    pass


class HyperlinkedModelSerializer(
        IncludedResourcesValidationMixin, SparseFieldsetsMixin, HyperlinkedModelSerializer,
        metaclass=SerializerMetaclass
):
    """
    A type of `ModelSerializer` that uses hyperlinked relationships instead
    of primary key relationships. Specifically:

    * A 'url' field is included instead of the 'id' field.
    * Relationships to other instances are hyperlinks, instead of primary keys.

    Included Mixins:

    * A mixin class to enable sparse fieldsets is included
    * A mixin class to enable validation of included resources is included
    """


class ModelSerializer(IncludedResourcesValidationMixin, SparseFieldsetsMixin, ModelSerializer,
                      metaclass=SerializerMetaclass):
    """
    A `ModelSerializer` is just a regular `Serializer`, except that:

    * A set of default fields are automatically populated.
    * A set of default validators are automatically populated.
    * Default `.create()` and `.update()` implementations are provided.

    The process of automatically determining a set of serializer fields
    based on the model fields is reasonably complex, but you almost certainly
    don't need to dig into the implementation.

    If the `ModelSerializer` class *doesn't* generate the set of fields that
    you need you should either declare the extra/differing fields explicitly on
    the serializer class, or simply use a `Serializer` class.


    Included Mixins:

    * A mixin class to enable sparse fieldsets is included
    * A mixin class to enable validation of included resources is included
    """
    serializer_related_field = ResourceRelatedField

    def get_field_names(self, declared_fields, info):
        """
        We override the parent to omit explicity defined meta fields (such
        as SerializerMethodFields) from the list of declared fields
        """
        meta_fields = getattr(self.Meta, 'meta_fields', [])

        declared = OrderedDict()
        for field_name in set(declared_fields.keys()):
            field = declared_fields[field_name]
            if field_name not in meta_fields:
                declared[field_name] = field
        fields = super(ModelSerializer, self).get_field_names(declared, info)
        return list(fields) + list(getattr(self.Meta, 'meta_fields', list()))

    def to_representation(self, instance):
        """
        Object instance -> Dict of primitive datatypes.
        """
        ret = OrderedDict()
        readable_fields = [
            field for field in self.fields.values()
            if not field.write_only
        ]

        for field in readable_fields:
            try:
                field_representation = self._get_field_representation(field, instance)
                ret[field.field_name] = field_representation
            except SkipField:
                continue

        return ret

    def _get_field_representation(self, field, instance):
        request = self.context.get('request')
        is_included = field.source in get_included_resources(request)
        render_nested_as_attribute = json_api_settings.SERIALIZE_NESTED_SERIALIZERS_AS_ATTRIBUTE
        if not is_included and \
                isinstance(field, ModelSerializer) and \
                hasattr(instance, field.source + '_id') and \
                not render_nested_as_attribute:
            attribute = getattr(instance, field.source + '_id')

            if attribute is None:
                return None

            resource_type = get_resource_type_from_serializer(field)
            if resource_type:
                return OrderedDict([('type', resource_type), ('id', attribute)])

        attribute = field.get_attribute(instance)

        # We skip `to_representation` for `None` values so that fields do
        # not have to explicitly deal with that case.
        #
        # For related fields with `use_pk_only_optimization` we need to
        # resolve the pk value.
        check_for_none = attribute.pk if isinstance(attribute, PKOnlyObject) else attribute
        if check_for_none is None:
            return None
        else:
            return field.to_representation(attribute)


class PolymorphicSerializerMetaclass(SerializerMetaclass):
    """
    This metaclass ensures that the `polymorphic_serializers` is correctly defined on a
    `PolymorphicSerializer` class and make a cache of model/serializer/type mappings.
    """

    def __new__(cls, name, bases, attrs):
        new_class = super(PolymorphicSerializerMetaclass, cls).__new__(cls, name, bases, attrs)

        # Ensure initialization is only performed for subclasses of PolymorphicModelSerializer
        # (excluding PolymorphicModelSerializer class itself).
        parents = [b for b in bases if isinstance(b, PolymorphicSerializerMetaclass)]
        if not parents:
            return new_class

        polymorphic_serializers = getattr(new_class, 'polymorphic_serializers', None)
        if not polymorphic_serializers:
            raise NotImplementedError(
                "A PolymorphicModelSerializer must define a `polymorphic_serializers` attribute.")
        serializer_to_model = {
            serializer: serializer.Meta.model for serializer in polymorphic_serializers}
        model_to_serializer = {
            serializer.Meta.model: serializer for serializer in polymorphic_serializers}
        type_to_serializer = {
            get_resource_type_from_serializer(serializer): serializer for
            serializer in polymorphic_serializers}
        new_class._poly_serializer_model_map = serializer_to_model
        new_class._poly_model_serializer_map = model_to_serializer
        new_class._poly_type_serializer_map = type_to_serializer
        new_class._poly_force_type_resolution = True

        # Flag each linked polymorphic serializer to force type resolution based on instance
        for serializer in polymorphic_serializers:
            serializer._poly_force_type_resolution = True

        return new_class


class PolymorphicModelSerializer(ModelSerializer, metaclass=PolymorphicSerializerMetaclass):
    """
    A serializer for polymorphic models.
    Useful for "lazy" parent models. Leaves should be represented with a regular serializer.
    """
    def get_fields(self):
        """
        Return an exhaustive list of the polymorphic serializer fields.
        """
        if self.instance not in (None, []):
            if not isinstance(self.instance, QuerySet):
                serializer_class = self.get_polymorphic_serializer_for_instance(self.instance)
                return serializer_class(self.instance, context=self.context).get_fields()
            else:
                raise Exception("Cannot get fields from a polymorphic serializer given a queryset")
        return super(PolymorphicModelSerializer, self).get_fields()

    @classmethod
    def get_polymorphic_serializer_for_instance(cls, instance):
        """
        Return the polymorphic serializer associated with the given instance/model.
        Raise `NotImplementedError` if no serializer is found for the given model. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        try:
            return cls._poly_model_serializer_map[instance._meta.model]
        except KeyError:
            raise NotImplementedError(
                "No polymorphic serializer has been found for model {}".format(
                    instance._meta.model.__name__))

    @classmethod
    def get_polymorphic_model_for_serializer(cls, serializer):
        """
        Return the polymorphic model associated with the given serializer.
        Raise `NotImplementedError` if no model is found for the given serializer. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        try:
            return cls._poly_serializer_model_map[serializer]
        except KeyError:
            raise NotImplementedError(
                "No polymorphic model has been found for serializer {}".format(serializer.__name__))

    @classmethod
    def get_polymorphic_serializer_for_type(cls, obj_type):
        """
        Return the polymorphic serializer associated with the given type.
        Raise `NotImplementedError` if no serializer is found for the given type. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        try:
            return cls._poly_type_serializer_map[obj_type]
        except KeyError:
            raise NotImplementedError(
                "No polymorphic serializer has been found for type {}".format(obj_type))

    @classmethod
    def get_polymorphic_model_for_type(cls, obj_type):
        """
        Return the polymorphic model associated with the given type.
        Raise `NotImplementedError` if no model is found for the given type. This usually
        means that a serializer is missing in the class's `polymorphic_serializers` attribute.
        """
        return cls.get_polymorphic_model_for_serializer(
            cls.get_polymorphic_serializer_for_type(obj_type))

    @classmethod
    def get_polymorphic_types(cls):
        """
        Return the list of accepted types.
        """
        return cls._poly_type_serializer_map.keys()

    def to_representation(self, instance):
        """
        Retrieve the appropriate polymorphic serializer and use this to handle representation.
        """
        serializer_class = self.get_polymorphic_serializer_for_instance(instance)
        return serializer_class(instance, context=self.context).to_representation(instance)

    def to_internal_value(self, data):
        """
        Ensure that the given type is one of the expected polymorphic types, then retrieve the
        appropriate polymorphic serializer and use this to handle internal value.
        """
        received_type = data.get('type')
        expected_types = self.get_polymorphic_types()
        if received_type not in expected_types:
            raise Conflict(
                'Incorrect relation type. Expected on of [{expected_types}], '
                'received {received_type}.'.format(
                    expected_types=', '.join(expected_types), received_type=received_type))
        serializer_class = self.get_polymorphic_serializer_for_type(received_type)
        self.__class__ = serializer_class
        return serializer_class(self.instance, data, context=self.context,
                                partial=self.partial).to_internal_value(data)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-01 08:34
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('example', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaggedItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('modified_at', models.DateTimeField(auto_now=True)),
                ('tag', models.SlugField()),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import pytest
from django.urls import reverse

pytestmark = pytest.mark.django_db


def test_included_data_on_list(multiple_entries, client):
    response = client.get(reverse("entry-list"), data={'include': 'comments', 'page[size]': 5})
    included = response.json().get('included')

    assert len(response.json()['data']) == len(multiple_entries), (
        'Incorrect entry count'
    )
    assert [x.get('type') for x in included] == ['comments', 'comments'], (
        'List included types are incorrect'
    )

    comment_count = len([resource for resource in included if resource["type"] == "comments"])
    expected_comment_count = sum([entry.comments.count() for entry in multiple_entries])
    assert comment_count == expected_comment_count, 'List comment count is incorrect'


def test_included_data_on_list_with_one_to_one_relations(multiple_entries, client):
    response = client.get(reverse("entry-list"),
                          data={'include': 'authors.bio.metadata', 'page[size]': 5})
    included = response.json().get('included')

    assert len(response.json()['data']) == len(multiple_entries), (
        'Incorrect entry count'
    )
    expected_include_types = [
        'authorBioMetadata', 'authorBioMetadata',
        'authorBios', 'authorBios',
        'authors', 'authors'
    ]
    include_types = [x.get('type') for x in included]
    assert include_types == expected_include_types, (
        'List included types are incorrect'
    )


def test_default_included_data_on_detail(single_entry, client):
    return test_included_data_on_detail(single_entry=single_entry, client=client, query='')


def test_included_data_on_detail(single_entry, client, query='?include=comments'):
    response = client.get(reverse("entry-detail", kwargs={'pk': single_entry.pk}) + query)
    included = response.json().get('included')

    assert [x.get('type') for x in included] == ['comments'], 'Detail included types are incorrect'

    comment_count = len([resource for resource in included if resource["type"] == "comments"])
    expected_comment_count = single_entry.comments.count()
    assert comment_count == expected_comment_count, 'Detail comment count is incorrect'


def test_dynamic_related_data_is_included(single_entry, entry_factory, client):
    entry_factory()
    response = client.get(
        reverse("entry-detail", kwargs={'pk': single_entry.pk}) + '?include=featured'
    )
    included = response.json().get('included')

    assert [x.get('type') for x in included] == ['entries'], 'Dynamic included types are incorrect'
    assert len(included) == 1, 'The dynamically included blog entries are of an incorrect count'


def test_dynamic_many_related_data_is_included(single_entry, entry_factory, client):
    entry_factory()
    response = client.get(
        reverse("entry-detail", kwargs={'pk': single_entry.pk}) + '?include=suggested'
    )
    included = response.json().get('included')

    assert included
    assert [x.get('type') for x in included] == ['entries'], 'Dynamic included types are incorrect'


def test_missing_field_not_included(author_bio_factory, author_factory, client):
    # First author does not have a bio
    author = author_factory(bio=None)
    response = client.get(reverse('author-detail', args=[author.pk]) + '?include=bio')
    assert 'included' not in response.json()
    # Second author does
    author = author_factory()
    response = client.get(reverse('author-detail', args=[author.pk]) + '?include=bio')
    data = response.json()
    assert 'included' in data
    assert len(data['included']) == 1
    assert data['included'][0]['attributes']['body'] == author.bio.body


def test_deep_included_data_on_list(multiple_entries, client):
    response = client.get(reverse("entry-list") + '?include=comments,comments.author,'
                          'comments.author.bio,comments.writer&page[size]=5')
    included = response.json().get('included')

    assert len(response.json()['data']) == len(multiple_entries), (
        'Incorrect entry count'
    )
    assert [x.get('type') for x in included] == [
        'authorBios', 'authorBios', 'authors', 'authors',
        'comments', 'comments', 'writers', 'writers'
    ], 'List included types are incorrect'

    comment_count = len([resource for resource in included if resource["type"] == "comments"])
    expected_comment_count = sum([entry.comments.count() for entry in multiple_entries])
    assert comment_count == expected_comment_count, 'List comment count is incorrect'

    author_count = len([resource for resource in included if resource["type"] == "authors"])
    expected_author_count = sum(
        [entry.comments.filter(author__isnull=False).count() for entry in multiple_entries])
    assert author_count == expected_author_count, 'List author count is incorrect'

    author_bio_count = len([resource for resource in included if resource["type"] == "authorBios"])
    expected_author_bio_count = sum([entry.comments.filter(
        author__bio__isnull=False).count() for entry in multiple_entries])
    assert author_bio_count == expected_author_bio_count, 'List author bio count is incorrect'

    writer_count = len(
        [resource for resource in included if resource["type"] == "writers"]
    )
    expected_writer_count = sum(
        [entry.comments.filter(author__isnull=False).count() for entry in multiple_entries])
    assert writer_count == expected_writer_count, 'List writer count is incorrect'

    # Also include entry authors
    response = client.get(reverse("entry-list") + '?include=authors,comments,comments.author,'
                          'comments.author.bio&page[size]=5')
    included = response.json().get('included')

    assert len(response.json()['data']) == len(multiple_entries), (
        'Incorrect entry count'
    )
    assert [x.get('type') for x in included] == [
        'authorBios', 'authorBios', 'authors', 'authors', 'authors', 'authors',
        'comments', 'comments'], 'List included types are incorrect'

    author_count = len([resource for resource in included if resource["type"] == "authors"])
    expected_author_count = sum(
        [entry.authors.count() for entry in multiple_entries] +
        [entry.comments.filter(author__isnull=False).count() for entry in multiple_entries])
    assert author_count == expected_author_count, 'List author count is incorrect'


def test_deep_included_data_on_detail(single_entry, client):
    # Same test as in list but also ensures that intermediate resources (here comments' authors)
    # are returned along with the leaf nodes
    response = client.get(reverse("entry-detail", kwargs={'pk': single_entry.pk}) +
                          '?include=comments,comments.author.bio')
    included = response.json().get('included')

    assert [x.get('type') for x in included] == ['authorBios', 'authors', 'comments'], \
        'Detail included types are incorrect'

    comment_count = len([resource for resource in included if resource["type"] == "comments"])
    expected_comment_count = single_entry.comments.count()
    assert comment_count == expected_comment_count, 'Detail comment count is incorrect'

    author_bio_count = len([resource for resource in included if resource["type"] == "authorBios"])
    expected_author_bio_count = single_entry.comments.filter(author__bio__isnull=False).count()
    assert author_bio_count == expected_author_bio_count, 'Detail author bio count is incorrect'


def test_data_resource_not_included_again(single_comment, client):
    # This test makes sure that the resource which is in the data field is excluded
    # from the included field.
    response = client.get(reverse("comment-detail", kwargs={'pk': single_comment.pk}) +
                          '?include=entry.comments')

    included = response.json().get('included')

    included_comments = [resource for resource in included if resource["type"] == "comments"]
    assert single_comment.pk not in [int(x.get('id')) for x in included_comments], \
        "Resource of the data field duplicated in included"

    comment_count = len(included_comments)
    expected_comment_count = single_comment.entry.comments.count()
    # The comment in the data attribute must not be included again.
    expected_comment_count -= 1
    assert comment_count == expected_comment_count, "Comment count incorrect"



import pytest
from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse
from django.utils import encoding

from example.tests import TestBase


class ModelViewSetTests(TestBase):
    """
    Test usage with ModelViewSets, also tests pluralization, camelization,
    and underscore.

    [<RegexURLPattern user-list ^identities/$>,
    <RegexURLPattern user-detail ^identities/(?P<pk>[^/]+)/$>]
    """
    list_url = reverse('user-list')

    def setUp(self):
        super(ModelViewSetTests, self).setUp()
        self.detail_url = reverse('user-detail', kwargs={'pk': self.miles.pk})

    def test_key_in_list_result(self):
        """
        Ensure the result has a 'user' key since that is the name of the model
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES='dasherize'):
            response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

        user = get_user_model().objects.all()[0]
        expected = {
            'data': [
                {
                    'type': 'users',
                    'id': encoding.force_str(user.pk),
                    'attributes': {
                        'first-name': user.first_name,
                        'last-name': user.last_name,
                        'email': user.email
                    },
                }
            ],
            'links': {
                'first': 'http://testserver/identities?page%5Bnumber%5D=1',
                'last': 'http://testserver/identities?page%5Bnumber%5D=2',
                'next': 'http://testserver/identities?page%5Bnumber%5D=2',
                'prev': None
            },
            'meta': {
                'pagination': {
                    'page': 1,
                    'pages': 2,
                    'count': 2
                }
            }
        }

        assert expected == response.json()

    def test_page_two_in_list_result(self):
        """
        Ensure that the second page is reachable and is the correct data.
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES='dasherize'):
            response = self.client.get(self.list_url, {'page[number]': 2})
        self.assertEqual(response.status_code, 200)

        user = get_user_model().objects.all()[1]
        expected = {
            'data': [
                {
                    'type': 'users',
                    'id': encoding.force_str(user.pk),
                    'attributes': {
                        'first-name': user.first_name,
                        'last-name': user.last_name,
                        'email': user.email
                    },
                }
            ],
            'links': {
                'first': 'http://testserver/identities?page%5Bnumber%5D=1',
                'last': 'http://testserver/identities?page%5Bnumber%5D=2',
                'next': None,
                'prev': 'http://testserver/identities?page%5Bnumber%5D=1'
            },
            'meta': {
                'pagination': {
                    'page': 2,
                    'pages': 2,
                    'count': 2
                }
            }
        }

        assert expected == response.json()

    def test_page_range_in_list_result(self):
        """
        Ensure that the range of a page can be changed from the client,
        tests pluralization as two objects means it converts ``user`` to
        ``users``.
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES='dasherize'):
            response = self.client.get(self.list_url, {'page[size]': 2})
        self.assertEqual(response.status_code, 200)

        users = get_user_model().objects.all()
        expected = {
            'data': [
                {
                    'type': 'users',
                    'id': encoding.force_str(users[0].pk),
                    'attributes': {
                        'first-name': users[0].first_name,
                        'last-name': users[0].last_name,
                        'email': users[0].email
                    },
                },
                {
                    'type': 'users',
                    'id': encoding.force_str(users[1].pk),
                    'attributes': {
                        'first-name': users[1].first_name,
                        'last-name': users[1].last_name,
                        'email': users[1].email
                    },
                }
            ],
            'links': {
                'first': 'http://testserver/identities?page%5Bnumber%5D=1&page%5Bsize%5D=2',
                'last': 'http://testserver/identities?page%5Bnumber%5D=1&page%5Bsize%5D=2',
                'next': None,
                'prev': None
            },
            'meta': {
                'pagination': {
                    'page': 1,
                    'pages': 1,
                    'count': 2
                }
            }
        }

        assert expected == response.json()

    def test_key_in_detail_result(self):
        """
        Ensure the result has a 'user' key.
        """
        with override_settings(JSON_API_FORMAT_FIELD_NAMES='dasherize'):
            response = self.client.get(self.detail_url)
        self.assertEqual(response.status_code, 200)

        expected = {
            'data': {
                'type': 'users',
                'id': encoding.force_str(self.miles.pk),
                'attributes': {
                    'first-name': self.miles.first_name,
                    'last-name': self.miles.last_name,
                    'email': self.miles.email
                },
            }
        }

        assert expected == response.json()

    def test_patch_requires_id(self):
        """
        Verify that 'id' is required to be passed in an update request.
        """
        data = {
            'data': {
                'type': 'users',
                'attributes': {
                    'first-name': 'DifferentName'
                }
            }
        }

        response = self.client.patch(self.detail_url, data=data)

        self.assertEqual(response.status_code, 400)

    def test_patch_requires_correct_id(self):
        """
        Verify that 'id' is the same then in url
        """
        data = {
            'data': {
                'type': 'users',
                'id': self.miles.pk + 1,
                'attributes': {
                    'first-name': 'DifferentName'
                }
            }
        }

        response = self.client.patch(self.detail_url, data=data)

        self.assertEqual(response.status_code, 409)

    def test_key_in_post(self):
        """
        Ensure a key is in the post.
        """
        self.client.login(username='miles', password='pw')
        data = {
            'data': {
                'type': 'users',
                'id': encoding.force_str(self.miles.pk),
                'attributes': {
                    'first-name': self.miles.first_name,
                    'last-name': self.miles.last_name,
                    'email': 'miles@trumpet.org'
                },
            }
        }

        with override_settings(JSON_API_FORMAT_FIELD_NAMES='dasherize'):
            response = self.client.put(self.detail_url, data=data)

        assert data == response.json()

        # is it updated?
        self.assertEqual(
            get_user_model().objects.get(pk=self.miles.pk).email,
            'miles@trumpet.org')

    def test_404_error_pointer(self):
        self.client.login(username='miles', password='pw')
        not_found_url = reverse('user-detail', kwargs={'pk': 12345})
        errors = {
            'errors': [
                {'detail': 'Not found.', 'status': '404', 'code': 'not_found'}
            ]
        }

        response = self.client.get(not_found_url)
        assert 404 == response.status_code
        assert errors == response.json()


@pytest.mark.django_db
def test_patch_allow_field_type(author, author_type_factory, client):
    """
    Verify that type field may be updated.
    """
    author_type = author_type_factory()
    url = reverse('author-detail', args=[author.id])

    data = {
        'data': {
            'id': author.id,
            'type': 'authors',
            'relationships': {
                'data': {
                    'id': author_type.id,
                    'type': 'author-type'
                }
            }
        }
    }

    response = client.patch(url, data=data)

    assert response.status_code == 200



from unittest import mock

import pytest
from django.urls import reverse

pytestmark = pytest.mark.django_db


@mock.patch(
    'rest_framework_json_api.utils'
    '.get_default_included_resources_from_serializer',
    new=lambda s: [])
def test_multiple_entries_no_pagination(multiple_entries, client):

    expected = {
        "data": [
            {
                "type": "posts",
                "id": "1",
                "attributes":
                {
                    "headline": multiple_entries[0].headline,
                    "bodyText": multiple_entries[0].body_text,
                    "pubDate": None,
                    "modDate": None
                },
                "meta": {
                    "bodyFormat": "text"
                },
                "relationships":
                {
                    "blog": {
                        "data": {"type": "blogs", "id": "1"}
                    },
                    "blogHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/1/blog",
                            "self": "http://testserver/entries/1/relationships/blog_hyperlinked"
                        }
                    },
                    "authors": {
                        "meta": {"count": 1},
                        "data": [{"type": "authors", "id": "1"}]
                    },
                    "comments": {
                        "meta": {"count": 1},
                        "data": [{"type": "comments", "id": "1"}]
                    },
                    "commentsHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/1/comments",
                            "self": "http://testserver/entries/1/relationships/comments_hyperlinked"
                        }
                    },
                    "suggested": {
                        "data": [{"type": "entries", "id": "2"}],
                        "links": {
                            "related": "http://testserver/entries/1/suggested/",
                            "self": "http://testserver/entries/1/relationships/suggested"
                        }
                    },
                    "suggestedHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/1/suggested/",
                            "self": "http://testserver/entries/1"
                                    "/relationships/suggested_hyperlinked"
                        }
                    },
                    "featuredHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/1/featured",
                            "self": "http://testserver/entries/1/relationships/featured_hyperlinked"
                        }
                    },
                    'tags': {'data': [], 'meta': {'count': 0}},
                }
            },
            {
                "type": "posts",
                "id": "2",
                "attributes":
                {
                    "headline": multiple_entries[1].headline,
                    "bodyText": multiple_entries[1].body_text,
                    "pubDate": None,
                    "modDate": None
                },
                "meta": {
                    "bodyFormat": "text"
                },
                "relationships":
                {
                    "blog": {
                        "data": {"type": "blogs", "id": "2"}
                    },
                    "blogHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/2/blog",
                            "self": "http://testserver/entries/2/relationships/blog_hyperlinked",
                        }
                    },
                    "authors": {
                        "meta": {"count": 1},
                        "data": [{"type": "authors", "id": "2"}]
                    },
                    "comments": {
                        "meta": {"count": 1},
                        "data": [{"type": "comments", "id": "2"}]
                    },
                    "commentsHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/2/comments",
                            "self": "http://testserver/entries/2/relationships/comments_hyperlinked"
                        }
                    },
                    "suggested": {
                        "data": [{"type": "entries", "id": "1"}],
                        "links": {
                            "related": "http://testserver/entries/2/suggested/",
                            "self": "http://testserver/entries/2/relationships/suggested"
                        }
                    },
                    "suggestedHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/2/suggested/",
                            "self": "http://testserver/entries/2"
                                    "/relationships/suggested_hyperlinked"
                        }
                    },
                    "featuredHyperlinked": {
                        "links": {
                            "related": "http://testserver/entries/2/featured",
                            "self": "http://testserver/entries/2/relationships/featured_hyperlinked"
                        }
                    },
                    'tags': {'data': [], 'meta': {'count': 0}},
                }
            },
        ]
    }

    response = client.get(reverse("nopage-entry-list"))

    assert expected == response.json()
